from typing import Set, ClassVar

from slixmpp.xmlstream import ElementBase


class Search(ElementBase):
    namespace = "jabber:iq:search"
    name = "query"
    plugin_attrib = "search"
    interfaces: ClassVar[Set[str]] = set()



import unittest
from slixmpp.test import SlixTest
from slixmpp.plugins.xep_0059 import Set
from slixmpp.xmlstream import ET


class TestSetStanzas(SlixTest):

    def testSetFirstIndex(self):
        s = Set()
        s['first'] = 'id'
        s.set_first_index('10')
        self.check(s, """
          <set xmlns="http://jabber.org/protocol/rsm">
            <first index="10">id</first>
          </set>
        """)

    def testGetFirstIndex(self):
        xml_string = """
          <set xmlns="http://jabber.org/protocol/rsm">
            <first index="10">id</first>
          </set>
        """
        s = Set(ET.fromstring(xml_string))
        expected = '10'
        self.assertTrue(s['first_index'] == expected)

    def testDelFirstIndex(self):
        xml_string = """
          <set xmlns="http://jabber.org/protocol/rsm">
            <first index="10">id</first>
          </set>
        """
        s = Set(ET.fromstring(xml_string))
        del s['first_index']
        self.check(s, """
          <set xmlns="http://jabber.org/protocol/rsm">
            <first>id</first>
          </set>
        """)

    def testSetBefore(self):
        s = Set()
        s['before'] = True
        self.check(s, """
          <set xmlns="http://jabber.org/protocol/rsm">
            <before />
          </set>
        """)

    def testGetBefore(self):
        xml_string = """
          <set xmlns="http://jabber.org/protocol/rsm">
            <before />
          </set>
        """
        s = Set(ET.fromstring(xml_string))
        expected = True
        self.assertTrue(s['before'] == expected)

    def testGetBefore(self):
        xml_string = """
          <set xmlns="http://jabber.org/protocol/rsm">
            <before />
          </set>
        """
        s = Set(ET.fromstring(xml_string))
        del s['before']
        self.check(s, """
          <set xmlns="http://jabber.org/protocol/rsm">
          </set>
        """)

    def testSetBeforeVal(self):
        s = Set()
        s['before'] = 'id'
        self.check(s, """
          <set xmlns="http://jabber.org/protocol/rsm">
            <before>id</before>
          </set>
        """)

    def testGetBeforeVal(self):
        xml_string = """
          <set xmlns="http://jabber.org/protocol/rsm">
            <before>id</before>
          </set>
        """
        s = Set(ET.fromstring(xml_string))
        expected = 'id'
        self.assertTrue(s['before'] == expected)

    def testGetBeforeVal(self):
        xml_string = """
          <set xmlns="http://jabber.org/protocol/rsm">
            <before>id</before>
          </set>
        """
        s = Set(ET.fromstring(xml_string))
        del s['before']
        self.check(s, """
          <set xmlns="http://jabber.org/protocol/rsm">
          </set>
        """)


suite = unittest.TestLoader().loadTestsFromTestCase(TestSetStanzas)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011 Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import uuid
import logging
import hashlib

from slixmpp.jid import JID
from slixmpp.exceptions import IqError, IqTimeout
from slixmpp.stanza import Iq, StreamFeatures
from slixmpp.xmlstream import ElementBase, ET, register_stanza_plugin
from slixmpp.plugins import BasePlugin
from slixmpp.plugins.xep_0078 import stanza


log = logging.getLogger(__name__)


class XEP_0078(BasePlugin):

    """
    XEP-0078 NON-SASL Authentication

    This XEP is OBSOLETE in favor of using SASL, so DO NOT use this plugin
    unless you are forced to use an old XMPP server implementation.
    """

    name = 'xep_0078'
    description = 'XEP-0078: Non-SASL Authentication'
    dependencies = set()
    stanza = stanza
    default_config = {
        'order': 15
    }

    def plugin_init(self):
        self.xmpp.register_feature('auth',
                self._handle_auth,
                restart=False,
                order=self.order)

        self.xmpp.add_event_handler('legacy_protocol',
                self._handle_legacy_protocol)

        register_stanza_plugin(Iq, stanza.IqAuth)
        register_stanza_plugin(StreamFeatures, stanza.AuthFeature)

    def plugin_end(self):
        self.xmpp.del_event_handler('legacy_protocol',
                self._handle_legacy_protocol)
        self.xmpp.unregister_feature('auth', self.order)

    def _handle_auth(self, features):
        # If we can or have already authenticated with SASL, do nothing.
        if 'mechanisms' in features['features']:
            return False
        return self.authenticate()

    def _handle_legacy_protocol(self, event):
        self.authenticate()

    def authenticate(self):
        if self.xmpp.authenticated:
            return False

        log.debug("Starting jabber:iq:auth Authentication")

        # Step 1: Request the auth form
        iq = self.xmpp.Iq()
        iq['type'] = 'get'
        iq['to'] = self.xmpp.requested_jid.host
        iq['auth']['username'] = self.xmpp.requested_jid.user

        try:
            resp = iq.send()
        except IqError as err:
            log.info("Authentication failed: %s", err.iq['error']['condition'])
            self.xmpp.event('failed_auth')
            self.xmpp.disconnect()
            return True
        except IqTimeout:
            log.info("Authentication failed: %s", 'timeout')
            self.xmpp.event('failed_auth')
            self.xmpp.disconnect()
            return True

        # Step 2: Fill out auth form for either password or digest auth
        iq = self.xmpp.Iq()
        iq['type'] = 'set'
        iq['auth']['username'] = self.xmpp.requested_jid.user

        # A resource is required, so create a random one if necessary
        resource = self.xmpp.requested_jid.resource
        if not resource:
            resource = str(uuid.uuid4())

        iq['auth']['resource'] = resource

        if 'digest' in resp['auth']['fields']:
            log.debug('Authenticating via jabber:iq:auth Digest')
            stream_id = bytes(self.xmpp.stream_id, encoding='utf-8')
            password = bytes(self.xmpp.password, encoding='utf-8')

            digest = hashlib.sha1(b'%s%s' % (stream_id, password)).hexdigest()
            iq['auth']['digest'] = digest
        else:
            log.warning('Authenticating via jabber:iq:auth Plain.')
            iq['auth']['password'] = self.xmpp.password

        # Step 3: Send credentials
        try:
            result = iq.send()
        except IqError as err:
            log.info("Authentication failed")
            self.xmpp.event("failed_auth")
            self.xmpp.disconnect()
        except IqTimeout:
            log.info("Authentication failed")
            self.xmpp.event("failed_auth")
            self.xmpp.disconnect()

        self.xmpp.features.add('auth')

        self.xmpp.authenticated = True

        self.xmpp.boundjid = JID(self.xmpp.requested_jid)
        self.xmpp.boundjid.resource = resource
        self.xmpp.event('session_bind', self.xmpp.boundjid)

        log.debug("Established Session")
        self.xmpp.sessionstarted = True
        self.xmpp.event('session_start')

        return True



#!/usr/bin/env python3

# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011  Dann Martens
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

import asyncio
import logging
from argparse import ArgumentParser
from getpass import getpass
from slixmpp.jid import JID
from slixmpp.plugins.xep_0009.remote import Endpoint, remote, Remote

class Thermostat(Endpoint):

    def FQN(self):
        return 'thermostat'

    def __init__(self, initial_temperature):
        pass

    @remote
    def set_temperature(self, temperature):
        return NotImplemented

    @remote
    def get_temperature(self):
        return NotImplemented

    @remote(False)
    def release(self):
        return NotImplemented


async def main(jid: JID, password: str, target_jid: JID):

    session = await Remote.new_session(jid, password)
    thermostat = session.new_proxy(target_jid, Thermostat)

    print("Current temperature is %s" % await thermostat.get_temperature())

    await thermostat.set_temperature(20)

    print("Current temperature is %s" % await thermostat.get_temperature())

    session.close()

if __name__ == '__main__':
    # Setup the command line arguments.
    parser = ArgumentParser()

    # Output verbosity options.
    parser.add_argument("-q", "--quiet", help="set logging to ERROR",
                        action="store_const", dest="loglevel",
                        const=logging.ERROR, default=logging.INFO)
    parser.add_argument("-d", "--debug", help="set logging to DEBUG",
                        action="store_const", dest="loglevel",
                        const=logging.DEBUG, default=logging.INFO)

    # JID and password options.
    parser.add_argument("-j", "--jid", dest="jid",
                        help="JID to use")
    parser.add_argument("-p", "--password", dest="password",
                        help="password to use")
    parser.add_argument("-t", "--target-jid", dest="target_jid",
                        help="target JID to use")

    args = parser.parse_args()

    # Setup logging.
    logging.basicConfig(level=args.loglevel,
                        format='%(levelname)-8s %(message)s')

    if args.jid is None:
        args.jid = input("Username: ")
    if args.password is None:
        args.password = getpass("Password: ")
    if args.target_jid is None:
        args.target_jid = input("Target jid: ")

    asyncio.run(main(JID(args.jid), args.password, args.target_jid))



import unittest
from slixmpp import Iq, Message, JID
from slixmpp.test import SlixTest
from slixmpp.plugins.xep_0405 import stanza
from slixmpp.plugins.xep_0369 import stanza as mstanza
from slixmpp.plugins.xep_0405.mix_pam import BASE_NODES


class TestMIXPAMStanza(SlixTest):

    def setUp(self):
        stanza.register_plugins()
        mstanza.register_plugins()

    def testMIXPAMJoin(self):
        """Test that data is converted to base64"""
        iq = Iq()
        iq['type'] = 'set'
        iq['client_join']['channel'] = JID('mix@example.com')
        for node in BASE_NODES:
            sub = mstanza.Subscribe()
            sub['node'] = node
            iq['client_join']['mix_join'].append(sub)
        iq['client_join']['mix_join']['nick'] = 'Toto'

        self.check(iq, """
          <iq type="set">
              <client-join xmlns='urn:xmpp:mix:pam:2' channel='mix@example.com'>
                  <join xmlns='urn:xmpp:mix:core:1'>
                      <subscribe node='urn:xmpp:mix:nodes:messages'/>
                      <subscribe node='urn:xmpp:mix:nodes:participants'/>
                      <subscribe node='urn:xmpp:mix:nodes:info'/>
                      <nick>Toto</nick>
                  </join>
              </client-join>
          </iq>
        """)


    def testMIXPAMLeave(self):
        iq = Iq()
        iq['type'] = 'set'
        iq['client_leave']['channel'] = JID('mix@example.com')
        iq['client_leave'].enable('mix_leave')

        self.check(iq, """
          <iq type="set">
              <client-leave xmlns='urn:xmpp:mix:pam:2' channel='mix@example.com'>
                  <leave xmlns='urn:xmpp:mix:core:1'/>
              </client-leave>
          </iq>
        """)


suite = unittest.TestLoader().loadTestsFromTestCase(TestMIXPAMStanza)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging

from asyncio import Future
from typing import (
    List,
    Optional,
    Set,
    Union,
)

from slixmpp.stanza import Iq
from slixmpp.plugins import BasePlugin
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.xmlstream import register_stanza_plugin, JID
from slixmpp.plugins.xep_0191 import stanza, Block, Unblock, BlockList


log = logging.getLogger(__name__)

BlockedJIDs = Union[
    JID,
    Set[JID],
    List[JID]
]


class XEP_0191(BasePlugin):

    name = 'xep_0191'
    description = 'XEP-0191: Blocking Command'
    dependencies = {'xep_0030'}
    stanza = stanza

    def plugin_init(self):
        register_stanza_plugin(Iq, BlockList)
        register_stanza_plugin(Iq, Block)
        register_stanza_plugin(Iq, Unblock)

        self.xmpp.register_handler(
                Callback('Blocked Contact',
                    StanzaPath('iq@type=set/block'),
                    self._handle_blocked))

        self.xmpp.register_handler(
                Callback('Unblocked Contact',
                    StanzaPath('iq@type=set/unblock'),
                    self._handle_unblocked))

    def plugin_end(self):
        self.xmpp.remove_handler('Blocked Contact')
        self.xmpp.remove_handler('Unblocked Contact')

    def get_blocked(self, ifrom: Optional[JID] = None, **iqkwargs) -> Future:
        """Get the list of blocked JIDs."""
        iq = self.xmpp.make_iq_get(ifrom=ifrom)
        iq.enable('blocklist')
        return iq.send(**iqkwargs)

    def block(self, jids: BlockedJIDs,
              ifrom: Optional[JID] = None, **iqkwargs) -> Future:
        """Block a JID or a list of JIDs.

        :param jids: JID(s) to block.
        """
        iq = self.xmpp.make_iq_set(ifrom=ifrom)
        if not isinstance(jids, (set, list)):
            jids = [jids]

        iq['block']['items'] = jids
        return iq.send(**iqkwargs)

    def unblock(self, jids: BlockedJIDs, ifrom: Optional[JID] = None, **iqkwargs) -> Future:
        """Unblock a JID or a list of JIDs.

        :param jids: JID(s) to unblock.
        """
        if jids is None:
            raise ValueError("jids cannot be empty.")
        iq = self.xmpp.make_iq_set(ifrom=ifrom)

        if not isinstance(jids, (set, list)):
            jids = [jids]

        iq['unblock']['items'] = jids
        return iq.send(**iqkwargs)

    def _handle_blocked(self, iq):
        self.xmpp.event('blocked', iq)

    def _handle_unblocked(self, iq):
        self.xmpp.event('unblocked', iq)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp import Iq, Message
from slixmpp.xmlstream import register_stanza_plugin, ElementBase, ET, JID
from slixmpp.plugins import xep_0004
from slixmpp.plugins.xep_0060.stanza.base import OptionalSetting


class Pubsub(ElementBase):
    namespace = 'http://jabber.org/protocol/pubsub'
    name = 'pubsub'
    plugin_attrib = name
    interfaces = set(tuple())


class Affiliations(ElementBase):
    namespace = 'http://jabber.org/protocol/pubsub'
    name = 'affiliations'
    plugin_attrib = name
    interfaces = {'node'}


class Affiliation(ElementBase):
    namespace = 'http://jabber.org/protocol/pubsub'
    name = 'affiliation'
    plugin_attrib = name
    interfaces = {'node', 'affiliation', 'jid'}

    def set_jid(self, value):
        self._set_attr('jid', str(value))

    def get_jid(self):
        return JID(self._get_attr('jid'))


class Subscription(ElementBase):
    namespace = 'http://jabber.org/protocol/pubsub'
    name = 'subscription'
    plugin_attrib = name
    interfaces = {'jid', 'node', 'subscription', 'subid'}

    def set_jid(self, value):
        self._set_attr('jid', str(value))

    def get_jid(self):
        return JID(self._get_attr('jid'))


class Subscriptions(ElementBase):
    namespace = 'http://jabber.org/protocol/pubsub'
    name = 'subscriptions'
    plugin_attrib = name
    interfaces = {'node'}


class SubscribeOptions(ElementBase, OptionalSetting):
    namespace = 'http://jabber.org/protocol/pubsub'
    name = 'subscribe-options'
    plugin_attrib = 'suboptions'
    interfaces = {'required'}


class Item(ElementBase):
    namespace = 'http://jabber.org/protocol/pubsub'
    name = 'item'
    plugin_attrib = name
    interfaces = {'id', 'payload'}

    def set_payload(self, value):
        del self['payload']
        if isinstance(value, ElementBase):
            if value.tag_name() in self.plugin_tag_map:
                self.init_plugin(value.plugin_attrib, existing_xml=value.xml)
            self.xml.append(value.xml)
        else:
            self.xml.append(value)

    def get_payload(self):
        children = list(self.xml)
        if len(children) > 0:
            return children[0]

    def del_payload(self):
        for child in self.xml:
            self.xml.remove(child)


class Items(ElementBase):
    namespace = 'http://jabber.org/protocol/pubsub'
    name = 'items'
    plugin_attrib = name
    interfaces = {'node', 'max_items'}

    def set_max_items(self, value):
        self._set_attr('max_items', str(value))


class Create(ElementBase):
    namespace = 'http://jabber.org/protocol/pubsub'
    name = 'create'
    plugin_attrib = name
    interfaces = {'node'}


class Default(ElementBase):
    namespace = 'http://jabber.org/protocol/pubsub'
    name = 'default'
    plugin_attrib = name
    interfaces = {'node', 'type'}

    def get_type(self):
        t = self._get_attr('type')
        if not t:
            return 'leaf'
        return t


class Publish(ElementBase):
    namespace = 'http://jabber.org/protocol/pubsub'
    name = 'publish'
    plugin_attrib = name
    interfaces = {'node'}


class Retract(ElementBase):
    namespace = 'http://jabber.org/protocol/pubsub'
    name = 'retract'
    plugin_attrib = name
    interfaces = {'node', 'notify'}

    def get_notify(self):
        notify = self._get_attr('notify')
        if notify in ('0', 'false'):
            return False
        elif notify in ('1', 'true'):
            return True
        return None

    def set_notify(self, value):
        del self['notify']
        if value is None:
            return
        elif value in (True, '1', 'true', 'True'):
            self._set_attr('notify', 'true')
        else:
            self._set_attr('notify', 'false')


class Unsubscribe(ElementBase):
    namespace = 'http://jabber.org/protocol/pubsub'
    name = 'unsubscribe'
    plugin_attrib = name
    interfaces = {'node', 'jid', 'subid'}

    def set_jid(self, value):
        self._set_attr('jid', str(value))

    def get_jid(self):
        return JID(self._get_attr('jid'))


class Subscribe(ElementBase):
    namespace = 'http://jabber.org/protocol/pubsub'
    name = 'subscribe'
    plugin_attrib = name
    interfaces = {'node', 'jid'}

    def set_jid(self, value):
        self._set_attr('jid', str(value))

    def get_jid(self):
        return JID(self._get_attr('jid'))


class Configure(ElementBase):
    namespace = 'http://jabber.org/protocol/pubsub'
    name = 'configure'
    plugin_attrib = name
    interfaces = {'node', 'type'}

    def getType(self):
        t = self._get_attr('type')
        if not t:
            t == 'leaf'
        return t


class Options(ElementBase):
    namespace = 'http://jabber.org/protocol/pubsub'
    name = 'options'
    plugin_attrib = name
    interfaces = {'jid', 'node', 'options'}

    def __init__(self, *args, **kwargs):
        ElementBase.__init__(self, *args, **kwargs)

    def get_options(self):
        config = self.xml.find('{jabber:x:data}x')
        form = xep_0004.Form(xml=config)
        return form

    def set_options(self, value):
        if isinstance(value, ElementBase):
            self.xml.append(value.xml)
        else:
            self.xml.append(value)
        return self

    def del_options(self):
        config = self.xml.find('{jabber:x:data}x')
        self.xml.remove(config)

    def set_jid(self, value):
        self._set_attr('jid', str(value))

    def get_jid(self):
        return JID(self._get_attr('jid'))


class PublishOptions(ElementBase):
    namespace = 'http://jabber.org/protocol/pubsub'
    name = 'publish-options'
    plugin_attrib = 'publish_options'
    interfaces = {'publish_options'}
    is_extension = True

    def get_publish_options(self):
        config = self.xml.find('{jabber:x:data}x')
        if config is None:
            return None
        form = xep_0004.Form(xml=config)
        return form

    def set_publish_options(self, value):
        if value is None:
            self.del_publish_options()
        else:
            if isinstance(value, ElementBase):
                self.xml.append(value.xml)
            else:
                self.xml.append(value)
        return self

    def del_publish_options(self):
        config = self.xml.find('{jabber:x:data}x')
        if config is not None:
            self.xml.remove(config)
        self.parent().xml.remove(self.xml)


register_stanza_plugin(Iq, Pubsub)
register_stanza_plugin(Pubsub, Affiliations)
register_stanza_plugin(Pubsub, Configure)
register_stanza_plugin(Pubsub, Create)
register_stanza_plugin(Pubsub, Default)
register_stanza_plugin(Pubsub, Items)
register_stanza_plugin(Pubsub, Options)
register_stanza_plugin(Pubsub, Publish)
register_stanza_plugin(Pubsub, PublishOptions)
register_stanza_plugin(Pubsub, Retract)
register_stanza_plugin(Pubsub, Subscribe)
register_stanza_plugin(Pubsub, Subscription)
register_stanza_plugin(Pubsub, Subscriptions)
register_stanza_plugin(Pubsub, Unsubscribe)
register_stanza_plugin(Affiliations, Affiliation, iterable=True)
register_stanza_plugin(Configure, xep_0004.Form)
register_stanza_plugin(Items, Item, iterable=True)
register_stanza_plugin(Publish, Item, iterable=True)
register_stanza_plugin(Retract, Item)
register_stanza_plugin(Subscribe, Options)
register_stanza_plugin(Subscription, SubscribeOptions)
register_stanza_plugin(Subscriptions, Subscription, iterable=True)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging

from slixmpp.stanza import StreamFeatures
from slixmpp.features.feature_rosterver import stanza
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.plugins.base import BasePlugin
from typing import ClassVar, Set


log = logging.getLogger(__name__)


class FeatureRosterVer(BasePlugin):

    name = 'feature_rosterver'
    description = 'RFC 6121: Stream Feature: Roster Versioning'
    dependences: ClassVar[Set[str]] = set()
    stanza = stanza

    def plugin_init(self):
        self.xmpp.register_feature('rosterver',
                self._handle_rosterver,
                restart=False,
                order=9000)

        register_stanza_plugin(StreamFeatures, stanza.RosterVer)

    def _handle_rosterver(self, features):
        """Enable using roster versioning.

        Arguments:
            features -- The stream features stanza.
        """
        log.debug("Enabling roster versioning.")
        self.xmpp.features.add('rosterver')




# slixmpp.util.stringprep_profiles
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# This module makes it easier to define profiles of stringprep,
# such as nodeprep and resourceprep for JID validation, and
# SASLprep for SASL.
# Part of Slixmpp: The Slick XMPP Library
# :copyright: (c) 2012 Nathanael C. Fritz, Lance J.T. Stout
# :license: MIT, see LICENSE for more details

from __future__ import unicode_literals

import stringprep
from unicodedata import ucd_3_2_0 as unicodedata

from slixmpp.util import unicode


class StringPrepError(UnicodeError):
    pass


def b1_mapping(char):
    """Map characters that are commonly mapped to nothing."""
    return '' if stringprep.in_table_b1(char) else None


def c12_mapping(char):
    """Map non-ASCII whitespace to spaces."""
    return ' ' if stringprep.in_table_c12(char) else None


def map_input(data, tables=None):
    """
    Each character in the input stream MUST be checked against
    a mapping table.
    """
    result = []
    for char in data:
        replacement = None

        for mapping in tables:
            replacement = mapping(char)
            if replacement is not None:
                break

        if replacement is None:
            replacement = char
        result.append(replacement)
    return ''.join(result)


def normalize(data, nfkc=True):
    """
    A profile can specify one of two options for Unicode normalization:
        - no normalization
        - Unicode normalization with form KC
    """
    if nfkc:
        data = unicodedata.normalize('NFKC', data)
    return data


def prohibit_output(data, tables=None):
    """
    Before the text can be emitted, it MUST be checked for prohibited
    code points.
    """
    for char in data:
        for check in tables:
            if check(char):
                raise StringPrepError("Prohibited code point: %s" % char)


def check_bidi(data):
    """
    1) The characters in section 5.8 MUST be prohibited.

    2) If a string contains any RandALCat character, the string MUST NOT
       contain any LCat character.

    3) If a string contains any RandALCat character, a RandALCat
       character MUST be the first character of the string, and a
       RandALCat character MUST be the last character of the string.
    """
    if not data:
        return data

    has_lcat = False
    has_randal = False

    for c in data:
        if stringprep.in_table_c8(c):
            raise StringPrepError("BIDI violation: seciton 6 (1)")
        if stringprep.in_table_d1(c):
            has_randal = True
        elif stringprep.in_table_d2(c):
            has_lcat = True

    if has_randal and has_lcat:
        raise StringPrepError("BIDI violation: section 6 (2)")

    first_randal = stringprep.in_table_d1(data[0])
    last_randal = stringprep.in_table_d1(data[-1])
    if has_randal and not (first_randal and last_randal):
        raise StringPrepError("BIDI violation: section 6 (3)")


def create(nfkc=True, bidi=True, mappings=None,
           prohibited=None, unassigned=None):
    """Create a profile of stringprep.

    :param bool nfkc:
        If `True`, perform NFKC Unicode normalization. Defaults to `True`.
    :param bool bidi:
        If `True`, perform bidirectional text checks. Defaults to `True`.
    :param list mappings:
        Optional list of functions for mapping characters to
        suitable replacements.
    :param list prohibited:
        Optional list of functions which check for the presence of
        prohibited characters.
    :param list unassigned:
        Optional list of functions for detecting the use of unassigned
        code points.

    :raises: StringPrepError
    :return: Unicode string of the resulting text passing the
             profile's requirements.
    """
    def profile(data, query=False):
        try:
            data = unicode(data)
        except UnicodeError:
            raise StringPrepError

        data = map_input(data, mappings)
        data = normalize(data, nfkc)
        prohibit_output(data, prohibited)
        if bidi:
            check_bidi(data)
        if query and unassigned:
            #check_unassigned(data, unassigned)
            raise StringPrepError('Query profile with unassigned data is unimplemented.')
        return data
    return profile



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import asyncio
import logging

from slixmpp.jid import JID
from slixmpp.stanza import Iq, StreamFeatures
from slixmpp.features.feature_bind import stanza
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.plugins import BasePlugin
from typing import ClassVar, Set


log = logging.getLogger(__name__)


class FeatureBind(BasePlugin):

    name = 'feature_bind'
    description = 'RFC 6120: Stream Feature: Resource Binding'
    dependencies: ClassVar[Set[str]] = set()
    stanza = stanza

    def plugin_init(self):
        self.xmpp.register_feature('bind',
                self._handle_bind_resource,
                restart=False,
                order=10000)

        register_stanza_plugin(Iq, stanza.Bind)
        register_stanza_plugin(StreamFeatures, stanza.Bind)

    async def _handle_bind_resource(self, features):
        """
        Handle requesting a specific resource.

        Arguments:
            features -- The stream features stanza.
        """
        log.debug("Requesting resource: %s", self.xmpp.requested_jid.resource)
        self.features = features
        iq = self.xmpp.Iq()
        iq['type'] = 'set'
        iq.enable('bind')
        if self.xmpp.requested_jid.resource:
            iq['bind']['resource'] = self.xmpp.requested_jid.resource

        await iq.send(callback=self._on_bind_response)

    def _on_bind_response(self, response):
        self.xmpp.boundjid = JID(response['bind']['jid'])
        self.xmpp.bound = True
        self.xmpp.event('session_bind', self.xmpp.boundjid)
        self.xmpp.session_bind_event.set()

        self.xmpp.features.add('bind')

        log.info("JID set to: %s", self.xmpp.boundjid.full)

        if 'session' not in self.features['features']:
            log.debug("Established Session")
            self.xmpp.sessionstarted = True
            self.xmpp.event('session_start')



#!/usr/bin/env python3

# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

import logging
from getpass import getpass
from argparse import ArgumentParser

import asyncio
import slixmpp


class CommandBot(slixmpp.ClientXMPP):

    """
    A simple Slixmpp bot that provides a basic
    adhoc command.
    """

    def __init__(self, jid, password):
        slixmpp.ClientXMPP.__init__(self, jid, password)

        # The session_start event will be triggered when
        # the bot establishes its connection with the server
        # and the XML streams are ready for use. We want to
        # listen for this event so that we we can initialize
        # our roster.
        self.add_event_handler("session_start", self.start)

    async def start(self, event):
        """
        Process the session_start event.

        Typical actions for the session_start event are
        requesting the roster and broadcasting an initial
        presence stanza.

        Arguments:
            event -- An empty dictionary. The session_start
                     event does not provide any additional
                     data.
        """
        self.send_presence()
        await self.get_roster()

        # We add the command after session_start has fired
        # to ensure that the correct full JID is used.

        # If using a component, may also pass jid keyword parameter.

        self['xep_0050'].add_command(node='greeting',
                                     name='Greeting',
                                     handler=self._handle_command)

    def _handle_command(self, iq, session):
        """
        Respond to the initial request for a command.

        Arguments:
            iq      -- The iq stanza containing the command request.
            session -- A dictionary of data relevant to the command
                       session. Additional, custom data may be saved
                       here to persist across handler callbacks.
        """
        form = self['xep_0004'].make_form('form', 'Greeting')
        form['instructions'] = 'Send a custom greeting to a JID'
        form.addField(var='greeting',
                      ftype='text-single',
                      label='Your greeting')

        session['payload'] = form
        session['next'] = self._handle_command_complete
        session['has_next'] = False

        # Other useful session values:
        # session['to']                    -- The JID that received the
        #                                     command request.
        # session['from']                  -- The JID that sent the
        #                                     command request.
        # session['has_next'] = True       -- There are more steps to complete
        # session['allow_complete'] = True -- Allow user to finish immediately
        #                                     and possibly skip steps
        # session['cancel'] = handler      -- Assign a handler for if the user
        #                                     cancels the command.
        # session['notes'] = [             -- Add informative notes about the
        #   ('info', 'Info message'),         command's results.
        #   ('warning', 'Warning message'),
        #   ('error', 'Error message')]

        return session

    def _handle_command_complete(self, payload, session):
        """
        Process a command result from the user.

        Arguments:
            payload -- Either a single item, such as a form, or a list
                       of items or forms if more than one form was
                       provided to the user. The payload may be any
                       stanza, such as jabber:x:oob for out of band
                       data, or jabber:x:data for typical data forms.
            session -- A dictionary of data relevant to the command
                       session. Additional, custom data may be saved
                       here to persist across handler callbacks.
        """

        # In this case (as is typical), the payload is a form
        form = payload

        greeting = form['values']['greeting']

        self.send_message(mto=session['from'],
                          mbody="%s, World!" % greeting,
                          mtype='chat')

        # Having no return statement is the same as unsetting the 'payload'
        # and 'next' session values and returning the session.

        # Unless it is the final step, always return the session dictionary.

        session['payload'] = None
        session['next'] = None

        return session


if __name__ == '__main__':
    # Setup the command line arguments.
    parser = ArgumentParser()

    # Output verbosity options.
    parser.add_argument("-q", "--quiet", help="set logging to ERROR",
                        action="store_const", dest="loglevel",
                        const=logging.ERROR, default=logging.INFO)
    parser.add_argument("-d", "--debug", help="set logging to DEBUG",
                        action="store_const", dest="loglevel",
                        const=logging.DEBUG, default=logging.INFO)

    # JID and password options.
    parser.add_argument("-j", "--jid", dest="jid",
                        help="JID to use")
    parser.add_argument("-p", "--password", dest="password",
                        help="password to use")

    args = parser.parse_args()

    # Setup logging.
    logging.basicConfig(level=args.loglevel,
                        format='%(levelname)-8s %(message)s')

    if args.jid is None:
        args.jid = input("Username: ")
    if args.password is None:
        args.password = getpass("Password: ")

    # Setup the CommandBot and register plugins. Note that while plugins may
    # have interdependencies, the order in which you register them does
    # not matter.
    xmpp = CommandBot(args.jid, args.password)
    xmpp.register_plugin('xep_0030') # Service Discovery
    xmpp.register_plugin('xep_0004') # Data Forms
    xmpp.register_plugin('xep_0050') # Adhoc Commands
    xmpp.register_plugin('xep_0199', {'keepalive': True, 'frequency':15})

    # Connect to the XMPP server and start processing XMPP stanzas.
    xmpp.connect()
    asyncio.get_event_loop().run_forever()



import logging

from slixmpp.plugins import BasePlugin

from . import stanza

log = logging.getLogger(__name__)


class XEP_0446(BasePlugin):

    """
    XEP-0446: File metadata element

    Minimum needed for xep 0447 (Stateless file sharing)
    """

    name = "xep_0446"
    description = "XEP-0446: File metadata element"
    stanza = stanza
    dependencies = {'xep_0300', 'xep_0264'}

    def plugin_init(self):
        stanza.register_plugins()



# -*- encoding: utf8 -*-
from __future__ import unicode_literals
import unittest
from slixmpp.test import SlixTest
from slixmpp import JID, InvalidJID


class TestJIDClass(SlixTest):

    """Verify that the JID class can parse and manipulate JIDs."""

    def testJIDFromFull(self):
        """Test using JID of the form 'user@server/resource/with/slashes'."""
        self.check_jid(JID('user@someserver/some/resource'),
                       'user',
                       'someserver',
                       'some/resource',
                       'user@someserver',
                       'user@someserver/some/resource',
                       'user@someserver/some/resource')

    def testJIDchange(self):
        """Test changing JID of the form 'user@server/resource/with/slashes'"""
        j = JID('user1@someserver1/some1/resource1')
        j.user = 'user'
        j.domain = 'someserver'
        j.resource = 'some/resource'
        self.check_jid(j,
                       'user',
                       'someserver',
                       'some/resource',
                       'user@someserver',
                       'user@someserver/some/resource',
                       'user@someserver/some/resource')

        j.node = None
        self.check_jid(j,
                       None,
                       'someserver',
                       'some/resource',
                       'someserver',
                       'someserver/some/resource',
                       'someserver/some/resource')

        j.resource = None
        self.check_jid(j,
                       None,
                       'someserver',
                       None,
                       'someserver',
                       'someserver',
                       'someserver')

    def testJIDaliases(self):
        """Test changing JID using aliases for domain."""
        j = JID('user@someserver/resource')
        j.server = 'anotherserver'
        self.check_jid(j, domain='anotherserver')
        j.host = 'yetanother'
        self.check_jid(j, domain='yetanother')

    def testJIDSetFullWithUser(self):
        """Test setting the full JID with a user portion."""
        j = JID('user@domain/resource')
        j.full = 'otheruser@otherdomain/otherresource'
        self.check_jid(j,
                       'otheruser',
                       'otherdomain',
                       'otherresource',
                       'otheruser@otherdomain',
                       'otheruser@otherdomain/otherresource',
                       'otheruser@otherdomain/otherresource')

    def testJIDFullNoUserWithResource(self):
        """
        Test setting the full JID without a user
        portion and with a resource.
        """
        j = JID('user@domain/resource')
        j.full = 'otherdomain/otherresource'
        self.check_jid(j,
                       '',
                       'otherdomain',
                       'otherresource',
                       'otherdomain',
                       'otherdomain/otherresource',
                       'otherdomain/otherresource')

    def testJIDFullNoUserNoResource(self):
        """
        Test setting the full JID without a user
        portion and without a resource.
        """
        j = JID('user@domain/resource')
        j.full = 'otherdomain'
        self.check_jid(j,
                       '',
                       'otherdomain',
                       '',
                       'otherdomain',
                       'otherdomain',
                       'otherdomain')

    def testJIDBareUser(self):
        """Test setting the bare JID with a user."""
        j = JID('user@domain/resource')
        j.bare = 'otheruser@otherdomain'
        self.check_jid(j,
                       'otheruser',
                       'otherdomain',
                       'resource',
                       'otheruser@otherdomain',
                       'otheruser@otherdomain/resource',
                       'otheruser@otherdomain/resource')

    def testJIDBareNoUser(self):
        """Test setting the bare JID without a user."""
        j = JID('user@domain/resource')
        j.bare = 'otherdomain'
        self.check_jid(j,
                       '',
                       'otherdomain',
                       'resource',
                       'otherdomain',
                       'otherdomain/resource',
                       'otherdomain/resource')

    def testJIDNoResource(self):
        """Test using JID of the form 'user@domain'."""
        self.check_jid(JID('user@someserver'),
                       'user',
                       'someserver',
                       '',
                       'user@someserver',
                       'user@someserver',
                       'user@someserver')

    def testJIDNoUser(self):
        """Test JID of the form 'component.domain.tld'."""
        self.check_jid(JID('component.someserver'),
                       '',
                       'component.someserver',
                       '',
                       'component.someserver',
                       'component.someserver',
                       'component.someserver')

    def testJIDEquality(self):
        """Test that JIDs with the same content are equal."""
        jid1 = JID('user@domain/resource')
        jid2 = JID('user@domain/resource')
        self.assertTrue(jid1 == jid2, "Same JIDs are not considered equal")
        self.assertFalse(jid1 != jid2, "Same JIDs are considered not equal")

    def testJIDInequality(self):
        jid1 = JID('user@domain/resource')
        jid2 = JID('otheruser@domain/resource')
        self.assertFalse(jid1 == jid2, "Different JIDs are considered equal")
        self.assertTrue(jid1 != jid2, "Different JIDs are considered equal")

    def testZeroLengthDomain(self):
        jid1 = JID('')
        jid2 = JID()
        self.assertTrue(jid1 == jid2, "Empty JIDs are not considered equal")
        self.assertTrue(jid1.domain == '', "Empty JID’s domain part not empty")
        self.assertTrue(jid1.full == '', "Empty JID’s full part not empty")

        self.assertRaises(InvalidJID, JID, 'user@')
        self.assertRaises(InvalidJID, JID, '/resource')
        self.assertRaises(InvalidJID, JID, 'user@/resource')

    def testZeroLengthLocalPart(self):
        self.assertRaises(InvalidJID, JID, '@test.com')
        self.assertRaises(InvalidJID, JID, '@test.com/resource')

    def testZeroLengthNodeDomain(self):
        self.assertRaises(InvalidJID, JID, '@/test.com')

    def testZeroLengthResource(self):
        self.assertRaises(InvalidJID, JID, 'test.com/')
        self.assertRaises(InvalidJID, JID, 'user@test.com/')

    def test1023LengthLocalPart(self):
        local = 'a' * 1023
        jid = JID('%s@test.com' % local)

    def test1023LengthResource(self):
        resource = 'r' * 1023
        jid = JID('test.com/%s' % resource)

    def test1024LengthDomain(self):
        domain = ('a.' * 509) + 'aa.com'
        self.assertRaises(InvalidJID, JID, 'user@%s/resource' % domain)
        self.assertRaises(InvalidJID, JID, 'user@%s' % domain)
        self.assertRaises(InvalidJID, JID, '%s/resource' % domain)
        self.assertRaises(InvalidJID, JID, domain)

    def test1024LengthLocalPart(self):
        local = 'a' * 1024
        self.assertRaises(InvalidJID, JID, '%s@test.com' % local)
        self.assertRaises(InvalidJID, JID, '%s@test.com/resource' % local)

    def test1024LengthResource(self):
        resource = 'r' * 1024
        self.assertRaises(InvalidJID, JID, 'test.com/%s' % resource)
        self.assertRaises(InvalidJID, JID, 'user@test.com/%s' % resource)

    def testTooLongDomainLabel(self):
        domain = ('a' * 64) + '.com'
        self.assertRaises(InvalidJID, JID, 'user@%s/resource' % domain)

    def testDomainEmptyLabel(self):
        domain = 'aaa..bbb.com'
        self.assertRaises(InvalidJID, JID, 'user@%s/resource' % domain)

    def testDomainIPv4(self):
        domain = '127.0.0.1'

        jid1 = JID('%s' % domain)
        jid2 = JID('user@%s' % domain)
        jid3 = JID('%s/resource' % domain)
        jid4 = JID('user@%s/resource' % domain)

    def testDomainIPv6(self):
        domain = '[::1]'

        jid1 = JID('%s' % domain)
        jid2 = JID('user@%s' % domain)
        jid3 = JID('%s/resource' % domain)
        jid4 = JID('user@%s/resource' % domain)

    def testDomainInvalidIPv6NoBrackets(self):
        domain = '::1'

        self.assertRaises(InvalidJID, JID, '%s' % domain)
        self.assertRaises(InvalidJID, JID, 'user@%s' % domain)
        self.assertRaises(InvalidJID, JID, '%s/resource' % domain)
        self.assertRaises(InvalidJID, JID, 'user@%s/resource' % domain)

    def testDomainInvalidIPv6MissingBracket(self):
        domain = '[::1'

        self.assertRaises(InvalidJID, JID, '%s' % domain)
        self.assertRaises(InvalidJID, JID, 'user@%s' % domain)
        self.assertRaises(InvalidJID, JID, '%s/resource' % domain)
        self.assertRaises(InvalidJID, JID, 'user@%s/resource' % domain)

    def testDomainInvalidIPv6WrongBracket(self):
        domain = '[::]1]'

        self.assertRaises(InvalidJID, JID, '%s' % domain)
        self.assertRaises(InvalidJID, JID, 'user@%s' % domain)
        self.assertRaises(InvalidJID, JID, '%s/resource' % domain)
        self.assertRaises(InvalidJID, JID, 'user@%s/resource' % domain)

    def testDomainWithPort(self):
        domain = 'example.com:5555'

        self.assertRaises(InvalidJID, JID, '%s' % domain)
        self.assertRaises(InvalidJID, JID, 'user@%s' % domain)
        self.assertRaises(InvalidJID, JID, '%s/resource' % domain)
        self.assertRaises(InvalidJID, JID, 'user@%s/resource' % domain)

    def testDomainWithTrailingDot(self):
        domain = 'example.com.'
        jid = JID('user@%s/resource' % domain)

        self.assertEqual(jid.domain, 'example.com')

    def testDomainWithDashes(self):
        domain = 'example.com-'
        self.assertRaises(InvalidJID, JID, 'user@%s/resource' % domain)

        domain = '-example.com'
        self.assertRaises(InvalidJID, JID, 'user@%s/resource' % domain)

    def testStartOrEndWithEscapedSpaces(self):
        local = ' foo'
        self.assertRaises(InvalidJID, JID, '%s@example.com' % local)

        local = 'bar '
        self.assertRaises(InvalidJID, JID, '%s@example.com' % local)

        # Need more input for these cases. A JID starting with \20 *is* valid
        # according to RFC 6122, but is not according to XEP-0106.
        #self.assertRaises(InvalidJID, JID, '%s@example.com' % '\\20foo2')
        #self.assertRaises(InvalidJID, JID, '%s@example.com' % 'bar2\\20')

    def testJapaneseLocalpart(self):
        jid = JID('あいうえお@example.com')
        self.assertEqual(jid.node, 'あいうえお')
        self.assertEqual(jid.domain, 'example.com')

    def testHash(self):
        jid = JID('あいうえお@example.com')
        self.assertEqual(hash(jid.full), hash(jid))
        jid = JID('toto@example.com/aaa')
        self.assertEqual(hash(jid.full), hash(jid))
        jid = JID('totoéà@example.com/aaa')
        self.assertEqual(hash(jid.full), hash(jid))


suite = unittest.TestLoader().loadTestsFromTestCase(TestJIDClass)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

import logging

from asyncio import Future
from typing import (
    Optional,
)

from slixmpp import Message
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import MatchXPath
from slixmpp.plugins.base import BasePlugin
from slixmpp.plugins.xep_0107 import stanza, UserMood


log = logging.getLogger(__name__)


class XEP_0107(BasePlugin):
    """
    XEP-0107: User Mood
    """

    name = 'xep_0107'
    description = 'XEP-0107: User Mood'
    dependencies = {'xep_0163'}
    stanza = stanza

    def plugin_init(self):
        register_stanza_plugin(Message, UserMood)

    def plugin_end(self):
        self.xmpp['xep_0030'].del_feature(feature=UserMood.namespace)
        self.xmpp['xep_0163'].remove_interest(UserMood.namespace)

    def session_bind(self, jid):
        self.xmpp['xep_0163'].register_pep('user_mood', UserMood)

    def publish_mood(self, value: Optional[str] = None, text: Optional[str] = None, **pubsubkwargs) -> Future:
        """
        Publish the user's current mood.

        :param value: The name of the mood to publish.
        :param text: Optional natural-language description or reason
                     for the mood.
        """
        mood = UserMood()
        mood['value'] = value
        mood['text'] = text
        return self.xmpp['xep_0163'].publish(
            mood,
            node=UserMood.namespace,
            **pubsubkwargs
        )

    def stop(self, **pubsubkwargs) -> Future:
        """
        Clear existing user mood information to stop notifications.
        """
        mood = UserMood()
        return self.xmpp['xep_0163'].publish(
            mood,
            node=UserMood.namespace,
            **pubsubkwargs
        )



#!/usr/bin/env python3

import logging
from getpass import getpass
from argparse import ArgumentParser

import asyncio
import slixmpp
from slixmpp.xmlstream import ET, tostring
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.xmlstream.handler import Callback


class PubsubEvents(slixmpp.ClientXMPP):

    def __init__(self, jid, password):
        super().__init__(jid, password)

        self.register_plugin('xep_0030')
        self.register_plugin('xep_0059')
        self.register_plugin('xep_0060')

        self.add_event_handler('session_start', self.start)

        # Some services may require configuration to allow
        # sending delete, configuration, or subscription  events.
        self.add_event_handler('pubsub_publish', self._publish)
        self.add_event_handler('pubsub_retract', self._retract)
        self.add_event_handler('pubsub_purge', self._purge)
        self.add_event_handler('pubsub_delete', self._delete)
        self.add_event_handler('pubsub_config', self._config)
        self.add_event_handler('pubsub_subscription', self._subscription)

        # Want to use nicer, more specific pubsub event names?
        # self['xep_0060'].map_node_event('node_name', 'event_prefix')
        # self.add_event_handler('event_prefix_publish', handler)
        # self.add_event_handler('event_prefix_retract', handler)
        # self.add_event_handler('event_prefix_purge', handler)
        # self.add_event_handler('event_prefix_delete', handler)

    async def start(self, event):
        await self.get_roster()
        self.send_presence()

    def _publish(self, msg):
        """Handle receiving a publish item event."""
        print('Published item %s to %s:' % (
            msg['pubsub_event']['items']['item']['id'],
            msg['pubsub_event']['items']['node']))
        data = msg['pubsub_event']['items']['item']['payload']
        if data is not None:
            print(tostring(data))
        else:
            print('No item content')

    def _retract(self, msg):
        """Handle receiving a retract item event."""
        print('Retracted item %s from %s' % (
            msg['pubsub_event']['items']['retract']['id'],
            msg['pubsub_event']['items']['node']))

    def _purge(self, msg):
        """Handle receiving a node purge event."""
        print('Purged all items from %s' % (
            msg['pubsub_event']['purge']['node']))

    def _delete(self, msg):
        """Handle receiving a node deletion event."""
        print('Deleted node %s' % (
           msg['pubsub_event']['delete']['node']))

    def _config(self, msg):
        """Handle receiving a node configuration event."""
        print('Configured node %s:' % (
            msg['pubsub_event']['configuration']['node']))
        print(msg['pubsub_event']['configuration']['form'])

    def _subscription(self, msg):
        """Handle receiving a node subscription event."""
        print('Subscription change for node %s:' % (
            msg['pubsub_event']['subscription']['node']))
        print(msg['pubsub_event']['subscription'])


if __name__ == '__main__':
    # Setup the command line arguments.
    parser = ArgumentParser()

    # Output verbosity options.
    parser.add_argument("-q", "--quiet", help="set logging to ERROR",
                        action="store_const", dest="loglevel",
                        const=logging.ERROR, default=logging.INFO)
    parser.add_argument("-d", "--debug", help="set logging to DEBUG",
                        action="store_const", dest="loglevel",
                        const=logging.DEBUG, default=logging.INFO)

    # JID and password options.
    parser.add_argument("-j", "--jid", dest="jid",
                        help="JID to use")
    parser.add_argument("-p", "--password", dest="password",
                        help="password to use")

    args = parser.parse_args()

    # Setup logging.
    logging.basicConfig(level=args.loglevel,
                        format='%(levelname)-8s %(message)s')

    if args.jid is None:
        args.jid = input("Username: ")
    if args.password is None:
        args.password = getpass("Password: ")

    logging.info("Run this in conjunction with the pubsub_client.py " + \
                 "example to watch events happen as you give commands.")

    # Setup the PubsubEvents listener
    xmpp = PubsubEvents(args.jid, args.password)

    # Connect to the XMPP server and start processing XMPP stanzas.
    xmpp.connect()
    asyncio.get_event_loop().run_forever()



import dataclasses
from collections import defaultdict
from enum import Enum


class RosterAccess(str, Enum):
    NONE = "none"
    GET = "get"
    SET = "set"
    BOTH = "both"


class MessagePermission(str, Enum):
    NONE = "none"
    OUTGOING = "outgoing"


class IqPermission(str, Enum):
    NONE = "none"
    GET = "get"
    SET = "set"
    BOTH = "both"


class PresencePermission(str, Enum):
    NONE = "none"
    MANAGED_ENTITY = "managed_entity"
    ROSTER = "roster"


@dataclasses.dataclass
class Permissions:
    roster = RosterAccess.NONE
    message = MessagePermission.NONE
    iq = defaultdict(lambda: IqPermission.NONE)
    presence = PresencePermission.NONE



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from typing import (
    Iterable,
    Optional,
    Set,
    Tuple,
)
from slixmpp import JID
from slixmpp.xmlstream import (
    ElementBase,
    ET,
    register_stanza_plugin,
)


class DiscoItem(ElementBase):
    name = 'item'
    namespace = 'http://jabber.org/protocol/disco#items'
    plugin_attrib = name
    interfaces = {'jid', 'node', 'name'}

    def get_node(self) -> Optional[str]:
        """Return the item's node name or ``None``."""
        return self._get_attr('node', None)

    def get_name(self) -> Optional[str]:
        """Return the item's human readable name, or ``None``."""
        return self._get_attr('name', None)


class DiscoItems(ElementBase):

    """
    Example disco#items stanzas:

    .. code-block:: xml

        <iq type="get">
          <query xmlns="http://jabber.org/protocol/disco#items" />
        </iq>

        <iq type="result">
          <query xmlns="http://jabber.org/protocol/disco#items">
            <item jid="chat.example.com"
                  node="xmppdev"
                  name="XMPP Dev" />
            <item jid="chat.example.com"
                  node="slixdev"
                  name="Slixmpp Dev" />
          </query>
        </iq>

    """

    name = 'query'
    namespace = 'http://jabber.org/protocol/disco#items'
    plugin_attrib = 'disco_items'
    #: Stanza Interface:
    #:
    #: - ``node``: The name of the node to either
    #:   query or return info from.
    #: - ``items``: A list of 3-tuples, where each tuple contains
    #:   the JID, node, and name of an item.
    #:
    interfaces = {'node', 'items'}

    # Cache items
    _items: Set[Tuple[JID, Optional[str]]]

    def setup(self, xml: Optional[ET.ElementTree] = None):
        """
        Populate the stanza object using an optional XML object.

        Overrides ElementBase.setup

        Caches item information.

        :param xml: Use an existing XML object for the stanza's values.
        """
        ElementBase.setup(self, xml)
        self._items = {item[0:2] for item in self['items']}

    def add_item(self, jid: JID, node: Optional[str] = None,
                 name: Optional[str] = None):
        """
        Add a new item element. Each item is required to have a
        JID, but may also specify a node value to reference
        non-addressable entitities.

        :param jid: The JID for the item.
        :param node: Optional additional information to reference
                     non-addressable items.
        :param name: Optional human readable name for the item.
        """
        if (jid, node) not in self._items:
            self._items.add((jid, node))
            item = DiscoItem(parent=self)
            item['jid'] = jid
            item['node'] = node
            item['name'] = name
            self.iterables.append(item)
            return True
        return False

    def del_item(self, jid: JID, node: Optional[str] = None) -> bool:
        """
        Remove a single item.

        :param jid: JID of the item to remove.
        :param node: Optional extra identifying information.
        """
        if (jid, node) in self._items:
            for item_xml in self.xml.findall('{%s}item' % self.namespace):
                item = (item_xml.attrib['jid'],
                        item_xml.attrib.get('node', None))
                if item == (jid, node):
                    self.xml.remove(item_xml)
                    return True
        return False

    def get_items(self) -> Set[DiscoItem]:
        """Return all items."""
        items = set()
        for item in self['substanzas']:
            if isinstance(item, DiscoItem):
                items.add((item['jid'], item['node'], item['name']))
        return items

    def set_items(self, items: Iterable[DiscoItem]):
        """
        Set or replace all items. The given items must be in a
        list or set where each item is a tuple of the form:

            (jid, node, name)

        :param items: A series of items in tuple format.
        """
        self.del_items()
        for item in items:
            jid, node, name = item
            self.add_item(jid, node, name)

    def del_items(self):
        """Remove all items."""
        self._items = set()
        items = [i for i in self.iterables if isinstance(i, DiscoItem)]
        for item in items:
            self.xml.remove(item.xml)
            self.iterables.remove(item)


register_stanza_plugin(DiscoItems, DiscoItem, iterable=True)



import asyncio
import unittest
from slixmpp.test.integration import SlixIntegration


class TestIBB(SlixIntegration):
    async def asyncSetUp(self):
        await super().asyncSetUp()
        self.add_client(
            self.envjid('CI_ACCOUNT1'),
            self.envstr('CI_ACCOUNT1_PASSWORD'),
        )
        self.add_client(
            self.envjid('CI_ACCOUNT2'),
            self.envstr('CI_ACCOUNT2_PASSWORD'),
        )
        config = {'block_size': 256, 'auto_accept': True}
        self.register_plugins(['xep_0047'], [config])
        self.data = b'to' * 257
        await self.connect_clients()

    async def test_ibb(self):
        """Check we can send and receive data through ibb"""
        coro_in = self.clients[1].wait_until('ibb_stream_start')
        coro_out = self.clients[0]['xep_0047'].open_stream(
            self.clients[1].boundjid,
            sid='toto'
        )
        instream, outstream = await asyncio.gather(coro_in, coro_out)

        async def send_and_close():
            await outstream.sendall(self.data)
            await outstream.close()

        in_data, _ = await asyncio.gather(instream.gather(), send_and_close())

        self.assertEqual(self.data, in_data)


suite = unittest.TestLoader().loadTestsFromTestCase(TestIBB)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream import ElementBase


class VCardTempUpdate(ElementBase):
    name = 'x'
    namespace = 'vcard-temp:x:update'
    plugin_attrib = 'vcard_temp_update'
    interfaces = {'photo'}
    sub_interfaces = interfaces

    def set_photo(self, value):
        if value is not None:
            self._set_sub_text('photo', value, keep=True)
        else:
            self._del_sub('photo')

    def get_photo(self):
        photo = self.xml.find('{%s}photo' % self.namespace)
        if photo is None:
            return None
        return photo.text



#!/usr/bin/env python3

# Slixmpp: The Slick XMPP Library
# Copyright (C) 2015  Emmanuel Gil Peyrot
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

import logging
from getpass import getpass
from argparse import ArgumentParser

import asyncio
import slixmpp


class S5BReceiver(slixmpp.ClientXMPP):

    """
    A basic example of creating and using a SOCKS5 bytestream.
    """

    def __init__(self, jid, password, filename):
        slixmpp.ClientXMPP.__init__(self, jid, password)

        self.file = open(filename, 'wb')

        self.add_event_handler("socks5_connected", self.stream_opened)
        self.add_event_handler("socks5_data", self.stream_data)
        self.add_event_handler("socks5_closed", self.stream_closed)

    def stream_opened(self, sid):
        logging.info('Stream opened. %s', sid)

    def stream_data(self, data):
        self.file.write(data)

    def stream_closed(self, exception):
        logging.info('Stream closed. %s', exception)
        self.file.close()
        self.disconnect()

if __name__ == '__main__':
    # Setup the command line arguments.
    parser = ArgumentParser()

    # Output verbosity options.
    parser.add_argument("-q", "--quiet", help="set logging to ERROR",
                        action="store_const", dest="loglevel",
                        const=logging.ERROR, default=logging.INFO)
    parser.add_argument("-d", "--debug", help="set logging to DEBUG",
                        action="store_const", dest="loglevel",
                        const=logging.DEBUG, default=logging.INFO)

    # JID and password options.
    parser.add_argument("-j", "--jid", dest="jid",
                        help="JID to use")
    parser.add_argument("-p", "--password", dest="password",
                        help="password to use")
    parser.add_argument("-o", "--out", dest="filename",
                        help="file to save to")

    args = parser.parse_args()

    # Setup logging.
    logging.basicConfig(level=args.loglevel,
                        format='%(levelname)-8s %(message)s')

    if args.jid is None:
        args.jid = input("Username: ")
    if args.password is None:
        args.password = getpass("Password: ")
    if args.filename is None:
        args.filename = input("File path: ")

    # Setup the S5BReceiver and register plugins. Note that while plugins may
    # have interdependencies, the order in which you register them does
    # not matter.
    xmpp = S5BReceiver(args.jid, args.password, args.filename)
    xmpp.register_plugin('xep_0030') # Service Discovery
    xmpp.register_plugin('xep_0065', {
        'auto_accept': True
    }) # SOCKS5 Bytestreams

    # Connect to the XMPP server and start processing XMPP stanzas.
    xmpp.connect()
    asyncio.get_event_loop().run_until_complete(xmpp.disconnected)



import unittest
from slixmpp import Message, Presence, Iq, JID
from slixmpp.test import SlixTest
from slixmpp.plugins.xep_0045.stanza import (
    MUCPresence,
    MUCJoin,
    MUCMessage,
    MUCAdminQuery,
    MUCAdminItem,
    MUCHistory,
    MUCOwnerQuery,
    MUCOwnerDestroy,
    MUCStatus,
    MUCInvite,
    MUCDecline,
    MUCUserItem,
    MUCActor,
)
from slixmpp.xmlstream import register_stanza_plugin, ET


class TestMUC(SlixTest):

    def setUp(self):
        register_stanza_plugin(MUCPresence, MUCUserItem)
        register_stanza_plugin(MUCMessage, MUCUserItem)
        register_stanza_plugin(MUCUserItem, MUCActor)
        register_stanza_plugin(MUCMessage, MUCInvite)
        register_stanza_plugin(MUCMessage, MUCDecline)
        register_stanza_plugin(MUCMessage, MUCStatus)
        register_stanza_plugin(MUCPresence, MUCStatus)
        register_stanza_plugin(Presence, MUCPresence)
        register_stanza_plugin(Presence, MUCJoin)
        register_stanza_plugin(MUCJoin, MUCHistory)
        register_stanza_plugin(Message, MUCMessage)
        register_stanza_plugin(Iq, MUCAdminQuery)
        register_stanza_plugin(Iq, MUCOwnerQuery)
        register_stanza_plugin(MUCOwnerQuery, MUCOwnerDestroy)
        register_stanza_plugin(MUCAdminQuery, MUCAdminItem, iterable=True)


    def testPresence(self):
        presence = Presence()
        presence['from'] = JID('muc@service/nick')
        presence['muc']['item']['affiliation'] = 'member'
        presence['muc']['item']['role'] = 'participant'
        presence['muc']['status_codes'] = (100, 110, 210)

        self.check(presence, """
<presence from='muc@service/nick'>
  <x xmlns='http://jabber.org/protocol/muc#user'>
    <item affiliation='member' role='participant'/>
    <status code='100'/>
    <status code='110'/>
    <status code='210'/>
  </x>
</presence>
        """, use_values=False)

    def testPresenceReason(self):
        presence = Presence()
        presence['from'] = JID('muc@service/nick')
        presence['muc']['item']['affiliation'] = 'member'
        presence['muc']['item']['role'] = 'participant'
        presence['muc']['item']['reason'] = 'coucou'
        presence['muc']['item']['actor']['nick'] = 'JPR'

        self.check(presence, """
<presence from='muc@service/nick'>
  <x xmlns='http://jabber.org/protocol/muc#user'>
    <item affiliation='member' role='participant'>
      <actor nick="JPR"/>
      <reason>coucou</reason>
    </item>
  </x>
</presence>
        """, use_values=False)


    def testPresenceLegacy(self):
        presence = Presence()
        presence['from'] = JID('muc@service/nick')
        presence['muc']['affiliation'] = 'member'
        presence['muc']['role'] = 'participant'
        presence['muc']['status_codes'] = (100, 110, 210)

        self.check(presence, """
<presence from='muc@service/nick'>
  <x xmlns='http://jabber.org/protocol/muc#user'>
    <item affiliation='member' role='participant'/>
    <status code='100'/>
    <status code='110'/>
    <status code='210'/>
  </x>
</presence>
        """, use_values=False)


    def testInvite(self):
        message = Message()
        message['from'] = 'user@server'
        message['to'] = 'muc@service'
        message['muc']['invite']['to'] = JID('user2@server2')
        message['muc']['invite']['reason'] = 'Hey'

        self.check(message, """
<message
    from='user@server'
    to='muc@service'>
  <x xmlns='http://jabber.org/protocol/muc#user'>
    <invite to='user2@server2'>
      <reason>Hey</reason>
    </invite>
  </x>
</message>
        """, use_values=False)

    def testAdmin(self):
        iq = Iq()
        iq['id'] = '0'
        iq['type'] = 'set'
        iq.enable('mucadmin_query')
        items = [
            ('none', 'test@example/a'),
            ('owner', 'owner@example/a'),
        ]
        for aff, jid in items:
            item_el = MUCAdminItem()
            item_el['affiliation'] = aff
            item_el['jid'] = jid
            iq['mucadmin_query'].append(item_el)
        self.check(iq, """
<iq type='set' id='0'>
  <query xmlns='http://jabber.org/protocol/muc#admin'>
    <item jid='test@example/a'
          affiliation='none'/>
    <item jid='owner@example/a'
          affiliation='owner'/>
  </query>
</iq>
        """, use_values=False)

    def testSetAffiliation(self):
        iq = Iq()
        iq['type'] = 'set'
        iq['id'] = '1'
        iq['mucadmin_query']['item']['jid'] = JID('test@example.com')
        iq['mucadmin_query']['item']['affiliation'] = 'owner'
        self.check(iq, """
<iq type='set' id='1'>
  <query xmlns='http://jabber.org/protocol/muc#admin'>
    <item jid='test@example.com'
          affiliation='owner'/>
  </query>
</iq>
        """, use_values=False)


suite = unittest.TestLoader().loadTestsFromTestCase(TestMUC)



import unittest
from slixmpp import Iq, Message, JID
from slixmpp.test import SlixTest
from slixmpp.plugins.xep_0369 import stanza
from slixmpp.plugins.xep_0060 import stanza as pstanza
from slixmpp.plugins.xep_0369.mix_core import BASE_NODES


class TestMIXStanza(SlixTest):

    def setUp(self):
        stanza.register_plugins()

    def testMIXJoin(self):
        """Test that data is converted to base64"""
        iq = Iq()
        iq['type'] = 'set'
        for node in BASE_NODES:
            sub = stanza.Subscribe()
            sub['node'] = node
            iq['mix_join'].append(sub)
        iq['mix_join']['nick'] = 'Toto'

        self.check(iq, """
          <iq type="set">
              <join xmlns='urn:xmpp:mix:core:1'>
                  <subscribe node='urn:xmpp:mix:nodes:messages'/>
                  <subscribe node='urn:xmpp:mix:nodes:participants'/>
                  <subscribe node='urn:xmpp:mix:nodes:info'/>
                  <nick>Toto</nick>
              </join>
          </iq>
        """)

    def testMIXUpdateSub(self):
        iq = Iq()
        iq['type'] = 'set'
        iq.enable('mix_updatesub')
        sub = stanza.Subscribe()
        sub['node'] = 'urn:xmpp:mix:nodes:someothernode'
        iq['mix_updatesub'].append(sub)

        self.check(iq, """
          <iq type="set">
              <update-subscription xmlns='urn:xmpp:mix:core:1'>
                  <subscribe node='urn:xmpp:mix:nodes:someothernode'/>
              </update-subscription>
          </iq>
        """)

    def testMIXLeave(self):
        iq = Iq()
        iq['type'] = 'set'
        iq.enable('mix_leave')

        self.check(iq, """
          <iq type="set">
              <leave xmlns='urn:xmpp:mix:core:1'/>
          </iq>
        """)

    def testMIXSetNick(self):
        iq = Iq()
        iq['type'] = 'set'
        iq['mix_setnick']['nick'] = 'A nick'

        self.check(iq, """
          <iq type="set">
              <setnick xmlns='urn:xmpp:mix:core:1'>
                <nick>A nick</nick>
              </setnick>
          </iq>
        """)

    def testMIXMessage(self):
        msg = Message()
        msg['type'] = 'groupchat'
        msg['body'] = 'This is a message body'
        msg['mix']['nick'] = 'A nick'
        msg['mix']['jid'] = JID('toto@example.com')

        self.check(msg, """
            <message type="groupchat">
                <body>This is a message body</body>
                <mix xmlns="urn:xmpp:mix:core:1">
                    <nick>A nick</nick>
                    <jid>toto@example.com</jid>
                </mix>
            </message>
        """)

    def testMIXNewParticipant(self):
        msg = Message()
        msg['pubsub_event']['items']['node'] = 'urn:xmpp:mix:nodes:participants'
        item = pstanza.EventItem()
        item['id'] = '123456'
        item['mix_participant']['jid'] = JID('titi@example.com')
        item['mix_participant']['nick'] = 'Titi'
        msg['pubsub_event']['items'].append(item)

        self.check(msg, """
            <message>
              <event xmlns='http://jabber.org/protocol/pubsub#event'>
                <items node='urn:xmpp:mix:nodes:participants'>
                  <item id='123456'>
                    <participant xmlns='urn:xmpp:mix:core:1'>
                      <jid>titi@example.com</jid>
                      <nick>Titi</nick>
                    </participant>
                  </item>
                </items>
              </event>
            </message>
        """, use_values=False)


suite = unittest.TestLoader().loadTestsFromTestCase(TestMIXStanza)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging

from slixmpp.xmlstream import JID
from slixmpp.exceptions import IqError, IqTimeout


log = logging.getLogger(__name__)


class StaticCaps(object):

    """
    Extend the default StaticDisco implementation to provide
    support for extended identity information.
    """

    def __init__(self, xmpp, static):
        """
        Augment the default XEP-0030 static handler object.

        Arguments:
            static -- The default static XEP-0030 handler object.
        """
        self.xmpp = xmpp
        self.disco = self.xmpp['xep_0030']
        self.caps = self.xmpp['xep_0115']
        self.static = static
        self.jid_vers = {}

    async def supports(self, jid, node, ifrom, data):
        """
        Check if a JID supports a given feature.

        The data parameter may provide:
            feature  -- The feature to check for support.
            local    -- If true, then the query is for a JID/node
                        combination handled by this Slixmpp instance and
                        no stanzas need to be sent.
                        Otherwise, a disco stanza must be sent to the
                        remove JID to retrieve the info.
            cached   -- If true, then look for the disco info data from
                        the local cache system. If no results are found,
                        send the query as usual. The self.use_cache
                        setting must be set to true for this option to
                        be useful. If set to false, then the cache will
                        be skipped, even if a result has already been
                        cached. Defaults to false.
        """
        feature = data.get('feature', None)

        data = {'local': data.get('local', False),
                'cached': data.get('cached', True)}

        if not feature:
            return False

        if node in (None, ''):
            info = await self.caps.get_caps(jid)
            if info and feature in info['features']:
                return True

        try:
            info = await self.disco.get_info(jid=jid, node=node,
                                             ifrom=ifrom, **data)
            info = self.disco._wrap(ifrom, jid, info, True)
            return feature in info['disco_info']['features']
        except IqError:
            return False
        except IqTimeout:
            return None

    async def has_identity(self, jid, node, ifrom, data):
        """
        Check if a JID has a given identity.

        The data parameter may provide:
            category -- The category of the identity to check.
            itype    -- The type of the identity to check.
            lang     -- The language of the identity to check.
            local    -- If true, then the query is for a JID/node
                        combination handled by this Slixmpp instance and
                        no stanzas need to be sent.
                        Otherwise, a disco stanza must be sent to the
                        remove JID to retrieve the info.
            cached   -- If true, then look for the disco info data from
                        the local cache system. If no results are found,
                        send the query as usual. The self.use_cache
                        setting must be set to true for this option to
                        be useful. If set to false, then the cache will
                        be skipped, even if a result has already been
                        cached. Defaults to false.
        """
        identity = (data.get('category', None),
                    data.get('itype', None),
                    data.get('lang', None))

        data = {'local': data.get('local', False),
                'cached': data.get('cached', True)}

        trunc = lambda i: (i[0], i[1], i[2])

        if node in (None, ''):
            info = self.caps.get_caps(jid)
            if info and identity in map(trunc, info['identities']):
                return True

        try:
            info = await self.disco.get_info(jid=jid, node=node,
                                             ifrom=ifrom, **data)
            info = self.disco._wrap(ifrom, jid, info, True)
            return identity in map(trunc, info['disco_info']['identities'])
        except IqError:
            return False
        except IqTimeout:
            return None

    def cache_caps(self, jid, node, ifrom, data):
        verstring = data.get('verstring', None)
        info = data.get('info', None)
        if not verstring or not info:
            return
        self.caps.cache.store(verstring, info)

    def assign_verstring(self, jid, node, ifrom, data):
        if isinstance(jid, JID):
            jid = jid.full
        self.jid_vers[jid] = data.get('verstring', None)

    def get_verstring(self, jid, node, ifrom, data):
        return self.jid_vers.get(jid, None)

    async def get_caps(self, jid, node, ifrom, data):
        verstring = data.get('verstring', None)
        if verstring is None:
            return None
        return self.caps.cache.retrieve(verstring)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.xmlstream import ElementBase
from slixmpp.plugins import xep_0082


class Geoloc(ElementBase):

    """
    XMPP's <geoloc> stanza allows entities to know the current
    geographical or physical location of an entity. (XEP-0080: User Location)

    Example <geoloc> stanzas:

    ::

        <geoloc xmlns='http://jabber.org/protocol/geoloc'/>

        <geoloc xmlns='http://jabber.org/protocol/geoloc' xml:lang='en'>
          <accuracy>20</accuracy>
          <country>Italy</country>
          <lat>45.44</lat>
          <locality>Venice</locality>
          <lon>12.33</lon>
        </geoloc>

    Stanza Interface:
    ::

        accuracy    -- Horizontal GPS error in meters.
        alt         -- Altitude in meters above or below sea level.
        area        -- A named area such as a campus or neighborhood.
        bearing     -- GPS bearing (direction in which the entity is
                       heading to reach its next waypoint), measured in
                       decimal degrees relative to true north.
        building    -- A specific building on a street or in an area.
        country     -- The nation where the user is located.
        countrycode -- The ISO 3166 two-letter country code.
        datum       -- GPS datum.
        description -- A natural-language name for or description of
                       the location.
        error       -- Horizontal GPS error in arc minutes. Obsoleted by
                       the accuracy parameter.
        floor       -- A particular floor in a building.
        lat         -- Latitude in decimal degrees North.
        locality    -- A locality within the administrative region, such
                       as a town or city.
        lon         -- Longitude in decimal degrees East.
        postalcode  -- A code used for postal delivery.
        region      -- An administrative region of the nation, such
                       as a state or province.
        room        -- A particular room in a building.
        speed       -- The speed at which the entity is moving,
                       in meters per second.
        street      -- A thoroughfare within the locality, or a crossing
                       of two thoroughfares.
        text        -- A catch-all element that captures any other
                       information about the location.
        timestamp   -- UTC timestamp specifying the moment when the
                       reading was taken.
        uri         -- A URI or URL pointing to information about
                       the location.

    """

    namespace = 'http://jabber.org/protocol/geoloc'
    name = 'geoloc'
    interfaces = {'accuracy', 'alt', 'area', 'bearing', 'building',
                  'country', 'countrycode', 'datum', 'dscription',
                  'error', 'floor', 'lat', 'locality', 'lon',
                  'postalcode', 'region', 'room', 'speed', 'street',
                  'text', 'timestamp', 'uri'}
    sub_interfaces = interfaces
    plugin_attrib = name

    def exception(self, e):
        """
        Override exception passback for presence.
        """
        pass

    def set_accuracy(self, accuracy):
        """
        Set the value of the <accuracy> element.

        :param accuracy: Horizontal GPS error in meters
        """
        self._set_sub_text('accuracy', text=str(accuracy))
        return self

    def get_accuracy(self):
        """
        Return the value of the <accuracy> element as an integer.
        """
        p = self._get_sub_text('accuracy')
        if not p:
            return None
        else:
            try:
                return int(p)
            except ValueError:
                return None

    def set_alt(self, alt):
        """
        Set the value of the <alt> element.

        :param alt: Altitude in meters above or below sea level
        """
        self._set_sub_text('alt', text=str(alt))
        return self

    def get_alt(self):
        """
        Return the value of the <alt> element as an integer.
        """
        p = self._get_sub_text('alt')
        if not p:
            return None
        else:
            try:
                return int(p)
            except ValueError:
                return None

    def set_bearing(self, bearing):
        """
        Set the value of the <bearing> element.

        :param bearing: GPS bearing (direction in which the entity is heading
                       to reach its next waypoint), measured in decimal
                       degrees relative to true north
        """
        self._set_sub_text('bearing', text=str(bearing))
        return self

    def get_bearing(self):
        """
        Return the value of the <bearing> element as a float.
        """
        p = self._get_sub_text('bearing')
        if not p:
            return None
        else:
            try:
                return float(p)
            except ValueError:
                return None

    def set_error(self, error):
        """
        Set the value of the <error> element.

        :param error: Horizontal GPS error in arc minutes; this
                      element is deprecated in favor of <accuracy/>
        """
        self._set_sub_text('error', text=str(error))
        return self

    def get_error(self):
        """
        Return the value of the <error> element as a float.
        """
        p = self._get_sub_text('error')
        if not p:
            return None
        else:
            try:
                return float(p)
            except ValueError:
                return None

    def set_lat(self, lat):
        """
        Set the value of the <lat> element.

        :param lat: Latitude in decimal degrees North
        """
        self._set_sub_text('lat', text=str(lat))
        return self

    def get_lat(self):
        """
        Return the value of the <lat> element as a float.
        """
        p = self._get_sub_text('lat')
        if not p:
            return None
        else:
            try:
                return float(p)
            except ValueError:
                return None

    def set_lon(self, lon):
        """
        Set the value of the <lon> element.

        :param lon: Longitude in decimal degrees East
        """
        self._set_sub_text('lon', text=str(lon))
        return self

    def get_lon(self):
        """
        Return the value of the <lon> element as a float.
        """
        p = self._get_sub_text('lon')
        if not p:
            return None
        else:
            try:
                return float(p)
            except ValueError:
                return None

    def set_speed(self, speed):
        """
        Set the value of the <speed> element.

        :param speed: The speed at which the entity is moving,
                      in meters per second
        """
        self._set_sub_text('speed', text=str(speed))
        return self

    def get_speed(self):
        """
        Return the value of the <speed> element as a float.
        """
        p = self._get_sub_text('speed')
        if not p:
            return None
        else:
            try:
                return float(p)
            except ValueError:
                return None

    def set_timestamp(self, timestamp):
        """
        Set the value of the <timestamp> element.

        :param timestamp: UTC timestamp specifying the moment when
                          the reading was taken
        """
        self._set_sub_text('timestamp', text=str(xep_0082.datetime(timestamp)))
        return self

    def get_timestamp(self):
        """
        Return the value of the <timestamp> element as a DateTime.
        """
        p = self._get_sub_text('timestamp')
        if not p:
            return None
        else:
            return xep_0082.datetime(p)



import asyncio
import unittest
from slixmpp.test.integration import SlixIntegration


class TestBOB(SlixIntegration):
    async def asyncSetUp(self):
        await super().asyncSetUp()
        self.add_client(
            self.envjid('CI_ACCOUNT1'),
            self.envstr('CI_ACCOUNT1_PASSWORD'),
        )
        self.add_client(
            self.envjid('CI_ACCOUNT2'),
            self.envstr('CI_ACCOUNT2_PASSWORD'),
        )
        self.register_plugins(['xep_0231'])
        self.data = b'to' * 257
        await self.connect_clients()

    async def test_bob(self):
        """Check we can send and receive a BOB."""
        cid = await self.clients[0]['xep_0231'].set_bob(
            self.data,
            'image/jpeg',
        )
        recv = await self.clients[1]['xep_0231'].get_bob(
            jid=self.clients[0].boundjid,
            cid=cid,
        )

        self.assertEqual(self.data, recv['bob']['data'])


suite = unittest.TestLoader().loadTestsFromTestCase(TestBOB)



import unittest
from slixmpp import Message
from slixmpp.test import SlixTest
from slixmpp.plugins.xep_0428 import stanza as fallback_stanza
from slixmpp.plugins.xep_0461 import stanza


class TestReply(SlixTest):
    def setUp(self):
        fallback_stanza.register_plugins()
        stanza.register_plugins()

    def testReply(self):
        message = Message()
        message["reply"]["id"] = "some-id"
        message["body"] = "some-body"

        self.check(
            message,
            """
            <message>
              <reply xmlns="urn:xmpp:reply:0" id="some-id" />
              <body>some-body</body>
            </message>
            """,
        )

    def testFallback(self):
        message = Message()
        message["body"] = "12345\nrealbody"
        message["fallback"]["for"] = "NS"
        message["fallback"]["body"]["start"] = 0
        message["fallback"]["body"]["end"] = 6

        self.check(
            message,
            """
            <message xmlns="jabber:client">
              <body>12345\nrealbody</body>
              <fallback xmlns='urn:xmpp:fallback:0' for='NS'>
                <body start="0" end="6" />
              </fallback>
            </message>
            """,
        )

        assert message["fallback"].get_stripped_body("NS") == "realbody"

    def testAddFallBackHelper(self):
        msg = Message()
        msg["body"] = "Great"
        msg["reply"].add_quoted_fallback("Anna wrote:\nHi, how are you?")
        self.check(
            msg,  # language=XML
            """
        <message xmlns="jabber:client" type="normal">
            <body>> Anna wrote:\n> Hi, how are you?\nGreat</body>
            <reply xmlns="urn:xmpp:reply:0" />
            <fallback xmlns="urn:xmpp:fallback:0" for="urn:xmpp:reply:0">
                <body start='0' end='33' />
            </fallback>
        </message>
            """
        )

    def testGetFallBackBody(self):
        body = "Anna wrote:\nHi, how are you?"
        quoted = "> Anna wrote:\n> Hi, how are you?\n"

        msg = Message()
        msg["body"] = "Great"
        msg["reply"].add_quoted_fallback(body)
        body2 = msg["reply"].get_fallback_body()
        self.assertTrue(body2 == quoted, body2)


suite = unittest.TestLoader().loadTestsFromTestCase(TestReply)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging

from slixmpp import Iq
from slixmpp.plugins import BasePlugin
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.plugins.xep_0257 import stanza, Certs
from slixmpp.plugins.xep_0257 import AppendCert, DisableCert, RevokeCert


log = logging.getLogger(__name__)


class XEP_0257(BasePlugin):

    name = 'xep_0257'
    description = 'XEP-0257: Client Certificate Management for SASL EXTERNAL'
    dependencies = {'xep_0030'}
    stanza = stanza

    def plugin_init(self):
        register_stanza_plugin(Iq, Certs)
        register_stanza_plugin(Iq, AppendCert)
        register_stanza_plugin(Iq, DisableCert)
        register_stanza_plugin(Iq, RevokeCert)

    def get_certs(self, ifrom=None, timeout=None, callback=None,
                        timeout_callback=None):
        iq = self.xmpp.Iq()
        iq['type'] = 'get'
        iq['from'] = ifrom
        iq.enable('sasl_certs')
        return iq.send(timeout=timeout, callback=callback,
                       timeout_callback=timeout_callback)

    def add_cert(self, name, cert, allow_management=True, ifrom=None,
                       timeout=None, callback=None, timeout_callback=None):
        iq = self.xmpp.Iq()
        iq['type'] = 'set'
        iq['from'] = ifrom
        iq['sasl_cert_append']['name'] = name
        iq['sasl_cert_append']['x509cert'] = cert
        iq['sasl_cert_append']['cert_management'] = allow_management
        return iq.send(timeout=timeout, callback=callback,
                       timeout_callback=timeout_callback)

    def disable_cert(self, name, ifrom=None, timeout=None, callback=None,
                                 timeout_callback=None):
        iq = self.xmpp.Iq()
        iq['type'] = 'set'
        iq['from'] = ifrom
        iq['sasl_cert_disable']['name'] = name
        return iq.send(timeout=timeout, callback=callback,
                       timeout_callback=timeout_callback)

    def revoke_cert(self, name, ifrom=None, timeout=None, callback=None,
                                timeout_callback=None):
        iq = self.xmpp.Iq()
        iq['type'] = 'set'
        iq['from'] = ifrom
        iq['sasl_cert_revoke']['name'] = name
        return iq.send(timeout=timeout, callback=callback,
                       timeout_callback=timeout_callback)



#!/usr/bin/env python3

# Slixmpp: The Slick XMPP Library
# Implementation of HTTP over XMPP transport
# http://xmpp.org/extensions/xep-0332.html
# Copyright (C) 2015 Riptide IO, sangeeth@riptideio.com
# This file is part of slixmpp.
# See the file LICENSE for copying permission.

import asyncio
from slixmpp import ClientXMPP

from argparse import ArgumentParser
import logging
import getpass


class HTTPOverXMPPClient(ClientXMPP):
    def __init__(self, jid, password):
        ClientXMPP.__init__(self, jid, password)
        self.register_plugin('xep_0332')    # HTTP over XMPP Transport
        self.add_event_handler(
            'session_start', self.session_start
        )
        self.add_event_handler('http_request', self.http_request_received)
        self.add_event_handler('http_response', self.http_response_received)

    def http_request_received(self, iq):
        pass

    def http_response_received(self, iq):
        print('HTTP Response Received : %s' % iq)
        print('From    : %s' %  iq['from'])
        print('To      : %s' % iq['to'])
        print('Type    : %s' % iq['type'])
        print('Headers : %s' % iq['resp']['headers'])
        print('Code    : %s' % iq['resp']['code'])
        print('Message : %s' % iq['resp']['message'])
        print('Data    : %s' % iq['resp']['data'])

    def session_start(self, event):
        # TODO: Fill in the blanks
        self['xep_0332'].send_request(
            to='?', method='?', resource='?', headers={}
        )
        self.disconnect()


if __name__ == '__main__':

    #
    # NOTE: To run this example, fill up the blanks in session_start() and
    #       use the following command.
    #
    # ./http_over_xmpp.py -J <jid> -P <pwd> -i <ip> -p <port> [-v]
    #

    parser = ArgumentParser()

    # Output verbosity options.
    parser.add_argument(
        '-v', '--verbose', help='set logging to DEBUG', action='store_const',
        dest='loglevel', const=logging.DEBUG, default=logging.ERROR
    )

    # JID and password options.
    parser.add_argument('-J', '--jid', dest='jid', help='JID')
    parser.add_argument('-P', '--password', dest='password', help='Password')

    # XMPP server ip and port options.
    parser.add_argument(
        '-i', '--ipaddr', dest='ipaddr',
        help='IP Address of the XMPP server', default=None
    )
    parser.add_argument(
        '-p', '--port', dest='port',
        help='Port of the XMPP server', default=None
    )

    args = parser.parse_args()

    # Setup logging.
    logging.basicConfig(level=args.loglevel,
                        format='%(levelname)-8s %(message)s')

    if args.jid is None:
        args.jid = input('Username: ')
    if args.password is None:
        args.password = getpass.getpass('Password: ')

    xmpp = HTTPOverXMPPClient(args.jid, args.password)
    xmpp.connect()
    asyncio.get_event_loop().run_forever()




# slixmpp.xmlstream.handler.collector
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Part of Slixmpp: The Slick XMPP Library
# :copyright: (c) 2012 Nathanael C. Fritz, Lance J.T. Stout
# :license: MIT, see LICENSE for more details
from __future__ import annotations

import logging
from typing import List, Optional, TYPE_CHECKING

from slixmpp.xmlstream.stanzabase import StanzaBase
from slixmpp.xmlstream.handler.base import BaseHandler
from slixmpp.xmlstream.matcher.base import MatcherBase

if TYPE_CHECKING:
    from slixmpp.xmlstream.xmlstream import XMLStream

log = logging.getLogger(__name__)


class Collector(BaseHandler):

    """
    The Collector handler allows for collecting a set of stanzas
    that match a given pattern. Unlike the Waiter handler, a
    Collector does not block execution, and will continue to
    accumulate matching stanzas until told to stop.

    :param string name: The name of the handler.
    :param matcher: A :class:`~slixmpp.xmlstream.matcher.base.MatcherBase`
                    derived object for matching stanza objects.
    :param stream: The :class:`~slixmpp.xmlstream.xmlstream.XMLStream`
                   instance this handler should monitor.
    """
    _stanzas: List[StanzaBase]

    def __init__(self, name: str, matcher: MatcherBase, stream: Optional[XMLStream] = None):
        BaseHandler.__init__(self, name, matcher, stream=stream)
        self._stanzas = []

    def prerun(self, payload: StanzaBase) -> None:
        """Store the matched stanza when received during processing.

        :param payload: The matched
            :class:`~slixmpp.xmlstream.stanzabase.StanzaBase` object.
        """
        self._stanzas.append(payload)

    def run(self, payload: StanzaBase) -> None:
        """Do not process this handler during the main event loop."""
        pass

    def stop(self) -> List[StanzaBase]:
        """
        Stop collection of matching stanzas, and return the ones that
        have been stored so far.
        """
        stream_ref = self.stream
        if stream_ref is None:
            raise ValueError('stop() called without a stream!')
        stream = stream_ref()
        if stream is None:
            raise ValueError('stop() called without a stream!')
        self._destroy = True
        stream.remove_handler(self.name)
        return self._stanzas




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2013 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging

from slixmpp import Iq, Message
from slixmpp.plugins import BasePlugin
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.xmlstream import register_stanza_plugin, JID
from slixmpp.plugins.xep_0096 import stanza, File


log = logging.getLogger(__name__)


class XEP_0096(BasePlugin):

    name = 'xep_0096'
    description = 'XEP-0096: SI File Transfer'
    dependencies = {'xep_0095'}
    stanza = stanza

    def plugin_init(self):
        register_stanza_plugin(self.xmpp['xep_0095'].stanza.SI, File)

        self.xmpp['xep_0095'].register_profile(File.namespace, self)

    def session_bind(self, jid):
        self.xmpp['xep_0030'].add_feature(File.namespace)

    def plugin_end(self):
        self.xmpp['xep_0030'].del_feature(feature=File.namespace)
        self.xmpp['xep_0095'].unregister_profile(File.namespace, self)

    def request_file_transfer(self, jid, sid=None, name=None, size=None,
                                    desc=None, hash=None, date=None,
                                    allow_ranged=False, mime_type=None,
                                    **iqargs):
        data = File()
        data['name'] = name
        data['size'] = size
        data['date'] = date
        data['desc'] = desc
        data['hash'] = hash
        if allow_ranged:
            data.enable('range')

        return self.xmpp['xep_0095'].offer(jid,
                sid=sid,
                mime_type=mime_type,
                profile=File.namespace,
                payload=data,
                **iqargs)



#!/usr/bin/env python3

# Slixmpp: The Slick XMPP Library
# Copyright (C) 2010  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.

import logging
from getpass import getpass
from argparse import ArgumentParser

import asyncio
import slixmpp
from slixmpp.exceptions import IqError, IqTimeout


class IBBSender(slixmpp.ClientXMPP):

    """
    A basic example of creating and using an in-band bytestream.
    """

    def __init__(self, jid, password, receiver, filename, use_messages=False):
        slixmpp.ClientXMPP.__init__(self, jid, password)

        self.receiver = receiver

        self.file = open(filename, 'rb')
        self.use_messages = use_messages

        # The session_start event will be triggered when
        # the bot establishes its connection with the server
        # and the XML streams are ready for use. We want to
        # listen for this event so that we we can initialize
        # our roster.
        self.add_event_handler("session_start", self.start)

    async def start(self, event):
        """
        Process the session_start event.

        Typical actions for the session_start event are
        requesting the roster and broadcasting an initial
        presence stanza.

        Arguments:
            event -- An empty dictionary. The session_start
                     event does not provide any additional
                     data.
        """
        self.send_presence()
        self.get_roster()

        try:
            # Open the IBB stream in which to write to.
            stream = await self['xep_0047'].open_stream(self.receiver, use_messages=self.use_messages)

            # If you want to send in-memory bytes, use stream.sendall() instead.
            await stream.sendfile(self.file, timeout=10)

            # And finally close the stream.
            await stream.close(timeout=10)
        except (IqError, IqTimeout):
            print('File transfer errored')
        else:
            print('File transfer finished')
        finally:
            self.file.close()
            self.disconnect()


if __name__ == '__main__':
    # Setup the command line arguments.
    parser = ArgumentParser()

    # Output verbosity options.
    parser.add_argument("-q", "--quiet", help="set logging to ERROR",
                        action="store_const", dest="loglevel",
                        const=logging.ERROR, default=logging.INFO)
    parser.add_argument("-d", "--debug", help="set logging to DEBUG",
                        action="store_const", dest="loglevel",
                        const=logging.DEBUG, default=logging.INFO)

    # JID and password options.
    parser.add_argument("-j", "--jid", dest="jid",
                        help="JID to use")
    parser.add_argument("-p", "--password", dest="password",
                        help="password to use")
    parser.add_argument("-r", "--receiver", dest="receiver",
                        help="JID of the receiver")
    parser.add_argument("-f", "--file", dest="filename",
                        help="file to send")
    parser.add_argument("-m", "--use-messages", action="store_true",
                        help="use messages instead of iqs for file transfer")

    args = parser.parse_args()

    # Setup logging.
    logging.basicConfig(level=args.loglevel,
                        format='%(levelname)-8s %(message)s')

    if args.jid is None:
        args.jid = input("Username: ")
    if args.password is None:
        args.password = getpass("Password: ")
    if args.receiver is None:
        args.receiver = input("Receiver: ")
    if args.filename is None:
        args.filename = input("File path: ")

    # Setup the IBBSender and register plugins. Note that while plugins may
    # have interdependencies, the order in which you register them does
    # not matter.
    xmpp = IBBSender(args.jid, args.password, args.receiver, args.filename, args.use_messages)
    xmpp.register_plugin('xep_0030') # Service Discovery
    xmpp.register_plugin('xep_0047') # In-band Bytestreams

    # Connect to the XMPP server and start processing XMPP stanzas.
    xmpp.connect()
    asyncio.get_event_loop().run_until_complete(xmpp.disconnected)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2017 Emmanuel Gil Peyrot <linkmauve@linkmauve.fr>
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0394.stanza import Markup, Span, BlockCode, List, Li, BlockQuote
from slixmpp.plugins.xep_0394.markup import XEP_0394


register_plugin(XEP_0394)



import unittest
from slixmpp.test.integration import SlixIntegration


class TestMarkers(SlixIntegration):
    async def asyncSetUp(self):
        self.add_client(
            self.envjid('CI_ACCOUNT1'),
            self.envstr('CI_ACCOUNT1_PASSWORD'),
        )
        self.add_client(
            self.envjid('CI_ACCOUNT2'),
            self.envstr('CI_ACCOUNT2_PASSWORD'),
        )
        self.register_plugins(['xep_0333'])
        await self.connect_clients()

    async def test_send_marker(self):
        """Send and receive a chat marker"""
        self.clients[0]['xep_0333'].send_marker(
            self.clients[1].boundjid.full,
            'toto',
            'displayed',
        )
        msg = await self.clients[1].wait_until('marker_displayed')

suite = unittest.TestLoader().loadTestsFromTestCase(TestMarkers)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permissio
from datetime import datetime
from typing import (
    Any,
    Iterable,
    List,
    Optional,
    Set,
    Union,
)

from slixmpp.stanza import Message
from slixmpp.jid import JID
from slixmpp.xmlstream import ElementBase, ET
from slixmpp.plugins import xep_0082


class MAM(ElementBase):
    """A MAM Query element.

    .. code-block:: xml

        <iq type='set' id='juliet1'>
          <query xmlns='urn:xmpp:mam:2'>
            <x xmlns='jabber:x:data' type='submit'>
              <field var='FORM_TYPE' type='hidden'>
                <value>urn:xmpp:mam:2</value>
              </field>
              <field var='with'>
                <value>juliet@capulet.lit</value>
              </field>
            </x>
          </query>
        </iq>

    """
    name = 'query'
    namespace = 'urn:xmpp:mam:2'
    plugin_attrib = 'mam'
    #: Available interfaces:
    #:
    #: - ``queryid``: The MAM query id
    #: - ``start`` and ``end``: Temporal boundaries of the query
    #: - ``with``: JID of the other entity the conversation is with
    #: - ``after_id``: Fetch stanzas after this specific ID
    #: - ``before_id``: Fetch stanzas before this specific ID
    #: - ``ids``: Fetch the stanzas matching those IDs
    #: - ``results``: pseudo-interface used to accumulate MAM results during
    #:   fetch, not relevant for the stanza itself.
    interfaces = {
        'queryid', 'start', 'end', 'with', 'results',
        'before_id', 'after_id', 'ids', 'flip_page',
    }
    sub_interfaces = {'start', 'end', 'with', 'before_id', 'after_id', 'ids',
                      'flip_page'}

    def setup(self, xml=None):
        ElementBase.setup(self, xml)
        self._results: List[Message] = []

    def _setup_form(self):
        found = self.xml.find(
                '{jabber:x:data}x/'
                '{jabber:x:data}field[@var="FORM_TYPE"]/'
                "{jabber:x:data}value[.='urn:xmpp:mam:2']"
        )
        if found is None:
            self['form']['type'] = 'submit'
            self['form'].add_field(
                var='FORM_TYPE', ftype='hidden', value='urn:xmpp:mam:2'
            )

    def get_fields(self):
        form = self.get_plugin('form', check=True)
        if not form:
            return {}
        return form.get_fields()

    def get_start(self) -> Optional[datetime]:
        fields = self.get_fields()
        field = fields.get('start')
        if field and field["value"]:
            return xep_0082.parse(field['value'])
        return None

    def set_start(self, value: Union[str, datetime]):
        self._setup_form()
        if isinstance(value, datetime):
            value = xep_0082.format_datetime(value)
        self.set_custom_field('start', value)

    def get_end(self) -> Optional[datetime]:
        fields = self.get_fields()
        field = fields.get('end')
        if field and field["value"]:
            return xep_0082.parse(field['value'])
        return None

    def set_end(self, value: Union[str, datetime]):
        if isinstance(value, datetime):
            value = xep_0082.format_datetime(value)
        self.set_custom_field('end', value)

    def get_with(self) -> Optional[JID]:
        fields = self.get_fields()
        field = fields.get('with')
        if field:
            return JID(field['value'])
        return None

    def set_with(self, value: JID):
        self.set_custom_field('with', value)

    def set_custom_field(self, fieldname: str, value: Any):
        self._setup_form()
        fields = self.get_fields()
        field = fields.get(fieldname)
        if field:
            field['value'] = str(value)
        else:
            field = self['form'].add_field(var=fieldname)
            field['value'] = str(value)

    def get_custom_field(self, fieldname: str) -> Optional[str]:
        fields = self.get_fields()
        field = fields.get(fieldname)
        if field:
            return field['value']
        return None

    def set_before_id(self, value: str):
        self.set_custom_field('before-id', value)

    def get_before_id(self):
        self.get_custom_field('before-id')

    def set_after_id(self, value: str):
        self.set_custom_field('after-id', value)

    def get_after_id(self):
        self.get_custom_field('after-id')

    def set_ids(self, value: List[str]):
        self._setup_form()
        fields = self.get_fields()
        field = fields.get('ids')
        if field:
            field['ids'] = value
        else:
            field = self['form'].add_field(var='ids')
            field['value'] = value

    def get_ids(self):
        self.get_custom_field('id')

    # The results interface is meant only as an easy
    # way to access the set of collected message responses
    # from the query.

    def get_results(self) -> List[Message]:
        return self._results

    def set_results(self, values: List[Message]):
        self._results = values

    def del_results(self):
        self._results = []

    def get_flip_page(self):
        return self.xml.find(f'{{{self.namespace}}}flip-page') is not None

class Fin(ElementBase):
    """A MAM fin element (end of query).

    .. code-block:: xml

        <iq type='result' id='juliet1'>
          <fin xmlns='urn:xmpp:mam:2'>
            <set xmlns='http://jabber.org/protocol/rsm'>
              <first index='0'>28482-98726-73623</first>
              <last>09af3-cc343-b409f</last>
            </set>
          </fin>
        </iq>

    """
    name = 'fin'
    namespace = 'urn:xmpp:mam:2'
    plugin_attrib = 'mam_fin'
    interfaces = {'results', 'stable', 'complete'}

    def setup(self, xml=None):
        ElementBase.setup(self, xml)
        self._results: List[Message] = []

    # The results interface is meant only as an easy
    # way to access the set of collected message responses
    # from the query.

    def get_results(self) -> List[Message]:
        return self._results

    def set_results(self, values: List[Message]):
        self._results = values

    def del_results(self):
        self._results = []


class Result(ElementBase):
    """A MAM result payload.

    .. code-block:: xml

        <message id='aeb213' to='juliet@capulet.lit/chamber'>
          <result xmlns='urn:xmpp:mam:2' queryid='f27' id='28482-98726-73623'>
            <forwarded xmlns='urn:xmpp:forward:0'>
              <delay xmlns='urn:xmpp:delay' stamp='2010-07-10T23:08:25Z'/>
              <message xmlns='jabber:client' from="witch@shakespeare.lit"
                       to="macbeth@shakespeare.lit">
                <body>Hail to thee</body>
              </message>
            </forwarded>
          </result>
        </message>
    """
    name = 'result'
    namespace = 'urn:xmpp:mam:2'
    plugin_attrib = 'mam_result'
    #: Available interfaces:
    #:
    #: - ``queryid``: MAM queryid
    #: - ``id``: ID of the result
    interfaces = {'queryid', 'id'}


class Metadata(ElementBase):
    """Element containing archive metadata

    .. code-block:: xml

        <iq type='result' id='jui8921rr9'>
          <metadata xmlns='urn:xmpp:mam:2'>
            <start id='YWxwaGEg' timestamp='2008-08-22T21:09:04Z' />
            <end id='b21lZ2Eg' timestamp='2020-04-20T14:34:21Z' />
          </metadata>
        </iq>

    """
    name = 'metadata'
    namespace = 'urn:xmpp:mam:2'
    plugin_attrib = 'mam_metadata'


class Start(ElementBase):
    """Metadata about the start of an archive.

    .. code-block:: xml

        <iq type='result' id='jui8921rr9'>
          <metadata xmlns='urn:xmpp:mam:2'>
            <start id='YWxwaGEg' timestamp='2008-08-22T21:09:04Z' />
            <end id='b21lZ2Eg' timestamp='2020-04-20T14:34:21Z' />
          </metadata>
        </iq>

    """
    name = 'start'
    namespace = 'urn:xmpp:mam:2'
    plugin_attrib = name
    #: Available interfaces:
    #:
    #: - ``id``: ID of the first message of the archive
    #: - ``timestamp`` (``datetime``): timestamp of the first message of the
    #:   archive
    interfaces = {'id', 'timestamp'}

    def get_timestamp(self) -> Optional[datetime]:
        """Get the timestamp.

        :returns: The timestamp.
        """
        stamp = self.xml.attrib.get('timestamp', None)
        if stamp is not None:
            return xep_0082.parse(stamp)
        return stamp

    def set_timestamp(self, value: Union[datetime, str]):
        """Set the timestamp.

        :param value: Value of the timestamp (either a datetime or a
                      XEP-0082 timestamp string.
        """
        if isinstance(value, str):
            value = xep_0082.parse(value)
        value = xep_0082.format_datetime(value)
        self.xml.attrib['timestamp'] = value


class End(ElementBase):
    """Metadata about the end of an archive.

    .. code-block:: xml

        <iq type='result' id='jui8921rr9'>
          <metadata xmlns='urn:xmpp:mam:2'>
            <start id='YWxwaGEg' timestamp='2008-08-22T21:09:04Z' />
            <end id='b21lZ2Eg' timestamp='2020-04-20T14:34:21Z' />
          </metadata>
        </iq>

    """
    name = 'end'
    namespace = 'urn:xmpp:mam:2'
    plugin_attrib = name
    #: Available interfaces:
    #:
    #: - ``id``: ID of the first message of the archive
    #: - ``timestamp`` (``datetime``): timestamp of the first message of the
    #:   archive
    interfaces = {'id', 'timestamp'}

    def get_timestamp(self) -> Optional[datetime]:
        """Get the timestamp.

        :returns: The timestamp.
        """
        stamp = self.xml.attrib.get('timestamp', None)
        if stamp is not None:
            return xep_0082.parse(stamp)
        return stamp

    def set_timestamp(self, value: Union[datetime, str]):
        """Set the timestamp.

        :param value: Value of the timestamp (either a datetime or a
                      XEP-0082 timestamp string.
        """
        if isinstance(value, str):
            value = xep_0082.parse(value)
        value = xep_0082.format_datetime(value)
        self.xml.attrib['timestamp'] = value



import unittest
import logging

from slixmpp import JID
from slixmpp.test import SlixTest

from slixmpp.plugins import xep_0100
from slixmpp.plugins.xep_0100 import LegacyError


class TestStreamGateway(SlixTest):
    def setUp(self):
        self.stream_start(
            mode="component",
            plugins=["xep_0077", "xep_0100"],
            jid="aim.shakespeare.lit",
            server="shakespeare.lit",
            plugin_config={
                "xep_0100": {"component_name": "AIM Gateway", "type": "aim"}
            },
        )
        self.xmpp._fix_error_ns()

    def next_sent(self):
        self.wait_for_send_queue()
        sent = self.xmpp.socket.next_sent(timeout=0.5)
        if sent is None:
            return None
        xml = self.parse_xml(sent)
        self.fix_namespaces(xml, "jabber:component:accept")
        sent = self.xmpp._build_stanza(xml, "jabber:component:accept")
        return sent

    def testDisco(self):
        # https://xmpp.org/extensions/xep-0100.html#example-3
        self.recv(
            """
            <iq type='get'
                from='romeo@montague.lit/orchard'
                to='aim.shakespeare.lit'
                id='disco1'>
            <query xmlns='http://jabber.org/protocol/disco#info'/>
            </iq>
            """
        )
        self.send(
            """
            <iq type="result"
                from="aim.shakespeare.lit"
                to="romeo@montague.lit/orchard"
                id="disco1">
            <query xmlns="http://jabber.org/protocol/disco#info">
                <identity category="gateway" type="aim" name="AIM Gateway" />
                <feature var="jabber:iq:register" />
                <feature var="jabber:x:data" />
                <feature var="jabber:iq:oob" />
                <feature var="jabber:x:oob" />
            </query>
            </iq>
            """
        )

    def testRegister(self):
        event_result = {}

        def legacy_login(iq):
            event_result["user"] = iq["from"]

        self.xmpp.add_event_handler("legacy_login", legacy_login)

        # Jabber User sends IQ-set qualified by the 'jabber:iq:register' namespace to Gateway,
        # containing information required to register.
        # https://xmpp.org/extensions/xep-0100.html#example-7
        self.recv(
            """
            <iq type='set'
                from='romeo@montague.lit/orchard'
                to='aim.shakespeare.lit'
                id='reg2'>
            <query xmlns='jabber:iq:register'>
                <username>RomeoMyRomeo</username>
                <password>ILoveJuliet</password>
            </query>
            </iq>
            """
        )
        # Gateway verifies that registration information provided by Jabber User is valid
        # (using whatever means appropriate for the Legacy Service) and informs Jabber User of success [A1].
        # https://xmpp.org/extensions/xep-0100.html#example-8
        self.send(
            """
            <iq type='result'
                from='aim.shakespeare.lit'
                to='romeo@montague.lit/orchard'
                id='reg2'/>
            """
        )
        # Gateway sends subscription request to Jabber User (i.e., by sending a presence stanza
        # of type "subscribe" to Jabber User's bare JID).
        # https://xmpp.org/extensions/xep-0100.html#example-11
        sent = self.next_sent()
        self.check(
            sent, "/presence@type=subscribe@from=aim.shakespeare.lit", "stanzapath"
        )
        self.assertTrue(
            sent["to"] == "romeo@montague.lit"
        )  # cannot use stanzapath because of @
        # Jabber User's client SHOULD approve the subscription request (i.e., by sending a presence stanza
        # of type "subscribed" to Gateway).
        self.recv(
            """
            <presence type='subscribed'
                      from='romeo@montague.lit'
                      to='aim.shakespeare.lit'/>
            """
        )
        # Jabber User sends subscription request to Gateway (i.e., by sending a presence stanza
        # of type "subscribe" to Gateway).
        self.recv(
            """
            <presence type='subscribe'
                      from='romeo@montague.lit'
                      to='aim.shakespeare.lit'/>
            """
        )
        # Gateway sends approves subscription request (i.e., by sending a presence stanza of type
        # "subscribed" to Jabber User's bare JID).
        sent = self.next_sent()
        self.check(
            sent, "/presence@type=subscribed@from=aim.shakespeare.lit", "stanzapath"
        )
        self.assertTrue(
            sent["to"] == "romeo@montague.lit"
        )  # cannot use stanzapath because of @
        self.assertTrue(
            self.xmpp.client_roster["romeo@montague.lit"]["subscription"] == "both"
        )
        self.recv(
            """
            <presence from='romeo@montague.lit/orchard'
                      to='aim.shakespeare.lit'/>
            """
        )
        self.assertTrue(event_result["user"] == "romeo@montague.lit/orchard")

    def testBadCredentials(self):
        def raise_v(*a, **kwa):
            raise ValueError("Not good")

        self.xmpp["xep_0077"].api.register(raise_v, "user_validate")
        self.recv(
            """
            <iq type='set'
                from='romeo@montague.lit/orchard'
                to='aim.shakespeare.lit'
                id='reg2'>
            <query xmlns='jabber:iq:register'>
                <username>RomeoMyRomeo</username>
                <password>ILoveJuliet</password>
            </query>
            </iq>
            """
        )
        self.send(
            """
            <iq type='error'
                from='aim.shakespeare.lit'
                to='romeo@montague.lit/orchard'
                id='reg2'>
            <query xmlns='jabber:iq:register'>
                <username>RomeoMyRomeo</username>
                <password>ILoveJuliet</password>
            </query>
            <error code='406' type='modify'>
                <not-acceptable
                    xmlns='urn:ietf:params:xml:ns:xmpp-stanzas'/>
                <text xmlns="urn:ietf:params:xml:ns:xmpp-stanzas">Not good</text>
            </error>
            </iq>
            """,
            use_values=False,
        )

    def testLogin(self):
        event_result = {}

        def legacy_login(presence):
            event_result["user"] = presence["from"]

        self.xmpp.add_event_handler("legacy_login", legacy_login)

        self.xmpp["xep_0077"].api["user_validate"](
            None,
            None,
            JID("romeo@montague.lit"),
            {"username": "RomeoMyRomeo", "password": "ILoveJuliet"},
        )

        # Jabber User sends available presence broadcast to Server or sends
        # directed presence to Gateway or a Legacy User.
        # https://xmpp.org/extensions/xep-0100.html#example-26
        self.recv(
            """
            <presence from='romeo@montague.lit/orchard'
                    to='juliet@aim.shakespeare.lit'/>
            <presence from='romeo@montague.lit/orchard'
                    to='aim.shakespeare.lit'/>
            """
        )
        # Gateway sends presence stanza to Jabber User expressing availability.
        # https://xmpp.org/extensions/xep-0100.html#example-27
        self.send(
            """
            <presence from='aim.shakespeare.lit'
                      to='romeo@montague.lit'>
            <priority>0</priority>
            </presence>
            """
        )
        self.assertTrue(event_result["user"] == "romeo@montague.lit/orchard")

    def testLogout(self):
        self.add_user()
        event_result = {}

        def legacy_logout(presence):
            event_result["user"] = presence["from"]

        self.xmpp.add_event_handler("legacy_logout", legacy_logout)
        # Jabber User sends available presence broadcast to Server or sends
        # directed presence to Gateway or a Legacy User.
        # https://xmpp.org/extensions/xep-0100.html#example-32
        self.recv(
            """
            <presence type='unavailable'
                    from='romeo@montague.lit/orchard'
                    to='aim.shakespeare.lit'/>
            """
        )
        # Gateway sends presence stanza of type "unavailable" to Jabber User.
        # https://xmpp.org/extensions/xep-0100.html#example-33
        self.send(
            """
            <presence type='unavailable'
                      from='aim.shakespeare.lit'
                      to='romeo@montague.lit/orchard'>
            <priority>0</priority>
            </presence>
            """
        )
        self.assertTrue(event_result["user"] == "romeo@montague.lit/orchard")

    def testAddContact(self):
        self.add_user()
        # Had to lowercase capuletnurse everywhere
        # Jabber User sends presence stanza of type "subscribe" to Legacy User.
        self.recv(
            """
            <presence type='subscribe'
                      from='romeo@montague.lit'
                      to='capuletnurse@aim.shakespeare.lit'/>
            """
        )
        # If Legacy User approves subscription request, Gateway sends presence stanza of
        # type "subscribed" to Jabber User on behalf of Legacy User. [A1]
        self.send(
            """
            <presence type='subscribed'
                      from='capuletnurse@aim.shakespeare.lit'
                      to='romeo@montague.lit'/>
            """
        )
        # Had to remove the resource here
        self.send(
            """
            <presence from='capuletnurse@aim.shakespeare.lit'
                      to='romeo@montague.lit'/>
            """
        )
        self.send(
            """
            <presence type='subscribe'
                      from='capuletnurse@aim.shakespeare.lit'
                      to='romeo@montague.lit'/>
            """
        )
        self.recv(
            """
            <presence type='subscribed'
                      from='romeo@montague.lit'
                      to='capuletnurse@aim.shakespeare.lit'/>
            """
        )

    def testAddContactFail(self):
        self.add_user()
        res = {}
        async def legacy_contact_add(jid, node, ifrom, contact_jid):
            res.update(**locals())
            raise LegacyError
        self.xmpp["xep_0100"].api.register(
            legacy_contact_add, "legacy_contact_add"
        )
        self.recv(
            """
            <presence type='subscribe'
                      from='romeo@montague.lit'
                      to='juliet@aim.shakespeare.lit'/>
            """
        )
        self.send(
            """
            <presence type='unsubscribed'
                      from='juliet@aim.shakespeare.lit'
                      to='romeo@montague.lit'/>
            """
        )
        self.assertTrue(res["ifrom"] == "romeo@montague.lit")
        self.assertTrue(res["contact_jid"] == "juliet@aim.shakespeare.lit")


    def testRemoveContact(self):
        self.add_user()
        result = {}
        # Jabber User sends IQ-set qualified by the 'jabber:iq:roster' namespace, containing subscription
        # attribute with value of "remove".
        async def legacy_contact_remove(jid, node, ifrom, contact_jid):
            result.update(**locals())

        self.xmpp["xep_0100"].api.register(
            legacy_contact_remove, "legacy_contact_remove"
        )

        # Jabber User sends IQ-set qualified by the 'jabber:iq:roster' namespace, containing subscription
        # attribute with value of "remove".
        self.recv(  # server sends this
            """
            <presence type='unsubscribe'
                      to='capuletnurse@aim.shakespeare.lit'
                      from='romeo@montague.lit'/>
            """
        )
        for ptype in "unsubscribe", "unsubscribed", "unavailable":
            self.send(  # server sends this
                f"""
                <presence type='{ptype}'
                        from='capuletnurse@aim.shakespeare.lit'
                        to='romeo@montague.lit'/>
                """
            )

        self.assertTrue(result["ifrom"] == "romeo@montague.lit")
        self.assertTrue(
            result["contact_jid"] == JID("CapuletNurse@aim.shakespeare.lit")
        )

    def testSendMessage(self):
        self.xmpp["xep_0100"].transform_legacy_message(
            jabber_user_jid="romeo@montague.lit",
            legacy_contact_id="juliet",
            body="Art thou not Romeo, and a Montague?",
        )
        self.send(
            """
            <message from='juliet@aim.shakespeare.lit'
                     to='romeo@montague.lit'>
            <body>Art thou not Romeo, and a Montague?</body>
            </message>
            """
        )

    def testLegacyMessage(self):
        self.add_user()
        result = {}

        def legacy_message(msg):
            result["msg"] = msg

        self.xmpp.add_event_handler("legacy_message", legacy_message)
        self.recv(
            """
            <message to='juliet@aim.shakespeare.lit'
                     from='romeo@montague.lit'>
            <body>Something shakespearian</body>
            </message>
            """
        )
        self.wait_for_send_queue()
        self.assertTrue(result["msg"]["from"] == "romeo@montague.lit")
        self.assertTrue(result["msg"]["to"] == "juliet@aim.shakespeare.lit")

    def testPluginEnd(self):
        exc = False
        try:
            self.xmpp.plugin.disable("xep_0100")
        except Exception:
            exc = True
        self.assertFalse(exc)

    def add_user(self):
        self.xmpp.loop.run_until_complete(
            self.xmpp["xep_0077"].api["user_validate"](
                None,
                None,
                JID("romeo@montague.lit"),
                {"username": "RomeoMyRomeo", "password": "ILoveJuliet"},
            )
        )

    # TODO: edit reg
    # TODO: unregister
    # TODO: login fails


logging.basicConfig(level=logging.DEBUG)
suite = unittest.TestLoader().loadTestsFromTestCase(TestStreamGateway)



import asyncio
import logging
from functools import partial
import typing

from slixmpp import Message, Iq, Presence, JID
from slixmpp.xmlstream.handler import Callback
from slixmpp.xmlstream.matcher import StanzaPath
from slixmpp.plugins import BasePlugin


class XEP_0100(BasePlugin):

    """
    XEP-0100: Gateway interaction

    Does not cover the deprecated Agent Information and 'jabber:iq:gateway' protocols

    Events registered by this plugin:

    - legacy_login: Jabber user got online or just registered
    - legacy_logout: Jabber user got offline or just unregistered
    - legacy_presence_unavailable: Jabber user sent an unavailable presence to a legacy contact
    - gateway_message: Jabber user sent a direct message to the gateway component
    - legacy_message: Jabber user sent a message to the legacy network


    Plugin Parameters:

    - `component_name`: (str) Name of the entity
    - `type`: (str) Type of the gateway identity. Should be the name of the legacy service
    - `needs_registration`: (bool) If set to True, messages received from unregistered users will
      not be transmitted to the legacy service

    API:

    - legacy_contact_add(jid, node, ifrom: JID, args: JID): Add contact on the legacy service.
      Should raise LegacyError if anything goes wrong in the process.
      `ifrom` is the gateway user's JID and `args` is the legacy contact's JID.
    - legacy_contact_remove(jid, node, ifrom: JID, args: JID): Remove a contact.

    """

    name = "xep_0100"
    description = "XEP-0100: Gateway interaction"
    dependencies = {
        "xep_0030",  # Service discovery
        "xep_0077",  # In band registration
    }

    default_config = {
        "component_name": "SliXMPP gateway",
        "type": "xmpp",
        "needs_registration": True,
    }

    def plugin_init(self):
        if not self.xmpp.is_component:
            log.error("Only components can be gateways, aborting plugin load")
            return

        self.xmpp["xep_0030"].add_identity(
            name=self.component_name, category="gateway", itype=self.type
        )

        self.api.register(self._legacy_contact_remove, "legacy_contact_remove")
        self.api.register(self._legacy_contact_add, "legacy_contact_add")

        # Without that BaseXMPP sends unsub/unavailable on sub requests and we don't want that
        self.xmpp.client_roster.auto_authorize = True
        self.xmpp.client_roster.auto_subscribe = False

        self.xmpp.add_event_handler("user_register", self.on_user_register)
        self.xmpp.add_event_handler("user_unregister", self.on_user_unregister)
        self.xmpp.add_event_handler("presence_available", self.on_presence_available)
        self.xmpp.add_event_handler(
            "presence_unavailable", self.on_presence_unavailable
        )
        self.xmpp.add_event_handler("presence_subscribe", self.on_presence_subscribe)
        self.xmpp.add_event_handler(
            "presence_unsubscribe", self.on_presence_unsubscribe
        )
        self.xmpp.add_event_handler("message", self.on_message)

    def plugin_end(self):
        if not self.xmpp.is_component:
            return

        self.xmpp.del_event_handler("user_register", self.on_user_register)
        self.xmpp.del_event_handler("user_unregister", self.on_user_unregister)
        self.xmpp.del_event_handler("presence_available", self.on_presence_available)
        self.xmpp.del_event_handler(
            "presence_unavailable", self.on_presence_unavailable
        )
        self.xmpp.del_event_handler("presence_subscribe", self.on_presence_subscribe)
        self.xmpp.del_event_handler("message", self.on_message)
        self.xmpp.del_event_handler(
            "presence_unsubscribe", self.on_presence_unsubscribe
        )

    async def get_user(self, stanza):
        return await self.xmpp["xep_0077"].api["user_get"](None, None, None, stanza)

    def send_presence(self, pto, ptype=None, pstatus=None, pfrom=None):
        self.xmpp.send_presence(
            pfrom=self.xmpp.boundjid.bare,
            ptype=ptype,
            pto=pto,
            pstatus=pstatus,
        )

    async def on_user_register(self, iq: Iq):
        user_jid = iq["from"]
        user = await self.get_user(iq)
        if user is None:  # This should not happen
            log.warning(f"{user_jid} has registered but cannot find them in user store")
        else:
            log.debug(f"Sending subscription request to {user_jid}")
            self.xmpp.client_roster.subscribe(user_jid)

    def on_user_unregister(self, iq: Iq):
        user_jid = iq["from"]
        log.debug(f"Sending subscription request to {user_jid}")
        self.xmpp.event("legacy_logout", iq)
        self.xmpp.client_roster.unsubscribe(iq["from"])
        self.xmpp.client_roster.remove(iq["from"])
        log.debug(f"roster: {self.xmpp.client_roster}")

    async def on_presence_available(self, presence: Presence):
        user_jid = presence["from"]
        user = await self.get_user(presence)
        if user is None:
            log.warning(
                f"{user_jid} has gotten online but cannot find them in user store"
            )
        else:
            self.xmpp.event("legacy_login", presence)
            log.debug(f"roster: {self.xmpp.client_roster}")
            self.send_presence(pto=user_jid.bare, ptype="available")

    async def on_presence_unavailable(self, presence: Presence):
        user_jid = presence["from"]
        user = await self.get_user(presence)
        if user is None:  # This should not happen
            log.warning(
                f"{user_jid} has gotten offline but but cannot find them in user store"
            )
            return

        if presence["to"] == self.xmpp.boundjid.bare:
            self.xmpp.event("legacy_logout", presence)
            self.send_presence(pto=user_jid, ptype="unavailable")
        else:
            self.xmpp.event("legacy_presence_unavailable", presence)

    async def _legacy_contact_add(self, jid, node, ifrom, contact_jid: JID):
        pass

    async def on_presence_subscribe(self, presence: Presence):
        user_jid = presence["from"]
        user = await self.get_user(presence)
        if user is None and self.needs_registration:
            return

        if presence["to"] == self.xmpp.boundjid.bare:
            return

        try:
            await self.api["legacy_contact_add"](
                ifrom=user_jid,
                args=presence["to"],
            )
        except LegacyError:
            self.xmpp.send_presence(
                pfrom=presence["to"],
                ptype="unsubscribed",
                pto=user_jid,
            )
            return
        self.xmpp.send_presence(
            pfrom=presence["to"],
            ptype="subscribed",
            pto=user_jid,
        )
        self.xmpp.send_presence(
            pfrom=presence["to"],
            pto=user_jid,
        )
        self.xmpp.send_presence(
            pfrom=presence["to"],
            ptype="subscribe",
            pto=user_jid,
        )  # TODO: handle resulting subscribed presences

    async def on_presence_unsubscribe(self, presence: Presence):
        if presence["to"] == self.xmpp.boundjid.bare:
            # should we trigger unregistering here?
            return

        user_jid = presence["from"]
        user = await self.get_user(presence)
        if user is None:
            log.debug("Received remove subscription from unregistered user")
            if self.needs_registration:
                return

        await self.api["legacy_contact_remove"](ifrom=user_jid, args=presence["to"])

        for ptype in "unsubscribe", "unsubscribed", "unavailable":
            self.xmpp.send_presence(
                pfrom=presence["to"],
                ptype=ptype,
                pto=user_jid,
            )

    async def _legacy_contact_remove(self, jid, node, ifrom, contact_jid: JID):
        pass

    async def on_message(self, msg: Message):
        if msg["type"] == "groupchat":
            return  # groupchat messages are out of scope of XEP-0100

        if msg["to"] == self.xmpp.boundjid.bare:
            # It may be useful to exchange direct messages with the component
            self.xmpp.event("gateway_message", msg)
            return

        if self.needs_registration and await self.get_user(msg) is None:
            return

        self.xmpp.event("legacy_message", msg)

    def transform_legacy_message(
        self,
        jabber_user_jid: typing.Union[JID, str],
        legacy_contact_id: str,
        body: str,
        mtype: typing.Optional[str] = None,
    ):
        """
        Transform a legacy message to an XMPP message
        """
        # Should escaping legacy IDs to valid JID local parts be handled here?
        # Maybe by internal API stuff?
        self.xmpp.send_message(
            mfrom=JID(f"{legacy_contact_id}@{self.xmpp.boundjid.bare}"),
            mto=JID(jabber_user_jid).bare,
            mbody=body,
            mtype=mtype,
        )


class LegacyError(Exception):
    pass


log = logging.getLogger(__name__)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp import Iq
from slixmpp.xmlstream import register_stanza_plugin
from slixmpp.plugins import BasePlugin
from slixmpp.plugins.xep_0016 import stanza
from slixmpp.plugins.xep_0016.stanza import Privacy, Item


class XEP_0016(BasePlugin):

    name = 'xep_0016'
    description = 'XEP-0016: Privacy Lists'
    dependencies = {'xep_0030'}
    stanza = stanza

    def plugin_init(self):
        register_stanza_plugin(Iq, Privacy)

    def plugin_end(self):
        self.xmpp['xep_0030'].del_feature(feature=Privacy.namespace)

    def session_bind(self, jid):
        self.xmpp['xep_0030'].add_feature(Privacy.namespace)

    def get_privacy_lists(self, timeout=None, callback=None,
                          timeout_callback=None):
        iq = self.xmpp.Iq()
        iq['type'] = 'get'
        iq.enable('privacy')
        iq.send(timeout=timeout, callback=callback,
                timeout_callback=timeout_callback)

    def get_list(self, name, timeout=None, callback=None, timeout_callback=None):
        iq = self.xmpp.Iq()
        iq['type'] = 'get'
        iq['privacy']['list']['name'] = name
        iq.send(timeout=timeout, callback=callback,
                timeout_callback=timeout_callback)

    def get_active(self, timeout=None, callback=None, timeout_callback=None):
        iq = self.xmpp.Iq()
        iq['type'] = 'get'
        iq['privacy'].enable('active')
        iq.send(timeout=timeout, callback=callback,
                       timeout_callback=timeout_callback)

    def get_default(self, timeout=None, callback=None,
                    timeout_callback=None):
        iq = self.xmpp.Iq()
        iq['type'] = 'get'
        iq['privacy'].enable('default')
        iq.send(timeout=timeout, callback=callback,
                timeout_callback=timeout_callback)

    def activate(self, name, timeout=None, callback=None,
                 timeout_callback=None):
        iq = self.xmpp.Iq()
        iq['type'] = 'set'
        iq['privacy']['active']['name'] = name
        iq.send(timeout=timeout, callback=callback,
                timeout_callback=timeout_callback)

    def deactivate(self, timeout=None, callback=None,
                   timeout_callback=None):
        iq = self.xmpp.Iq()
        iq['type'] = 'set'
        iq['privacy'].enable('active')
        iq.send(timeout=timeout, callback=callback,
                timeout_callback=timeout_callback)

    def make_default(self, name, timeout=None, callback=None,
                     timeout_callback=None):
        iq = self.xmpp.Iq()
        iq['type'] = 'set'
        iq['privacy']['default']['name'] = name
        iq.send(timeout=timeout, callback=callback,
                timeout_callback=timeout_callback)

    def remove_default(self, timeout=None, callback=None,
                       timeout_callback=None):
        iq = self.xmpp.Iq()
        iq['type'] = 'set'
        iq['privacy'].enable('default')
        iq.send(timeout=timeout, callback=callback,
                timeout_callback=timeout_callback)

    def edit_list(self, name, rules, timeout=None, callback=None,
                  timeout_callback=None):
        iq = self.xmpp.Iq()
        iq['type'] = 'set'
        iq['privacy']['list']['name'] = name
        priv_list = iq['privacy']['list']

        if not rules:
            rules = []

        for rule in rules:
            if isinstance(rule, Item):
                priv_list.append(rule)
                continue

            priv_list.add_item(
                rule['value'],
                rule['action'],
                rule['order'],
                itype=rule.get('type', None),
                iq=rule.get('iq', None),
                message=rule.get('message', None),
                presence_in=rule.get('presence_in',
                    rule.get('presence-in', None)),
                presence_out=rule.get('presence_out',
                    rule.get('presence-out', None)))

    def remove_list(self, name, timeout=None, callback=None,
                    timeout_callback=None):
        iq = self.xmpp.Iq()
        iq['type'] = 'set'
        iq['privacy']['list']['name'] = name
        iq.send(timeout=timeout, callback=callback,
                timeout_callback=timeout_callback)



import unittest
from slixmpp import Presence, JID
from slixmpp.test import SlixTest
from slixmpp.plugins.xep_0437 import stanza


class TestRAI(SlixTest):

    def setUp(self):
        stanza.register_plugins()

    def testResponse(self):
        presence = Presence()
        presence['rai']['activities'] = [
            JID('toto@titi'),
            JID('coucou@coucou'),
        ]
        self.check(presence, """
<presence>
  <rai xmlns="urn:xmpp:rai:0">
    <activity>toto@titi</activity>
    <activity>coucou@coucou</activity>
  </rai>
</presence>
        """, use_values=False)


suite = unittest.TestLoader().loadTestsFromTestCase(TestRAI)



import unittest
from slixmpp import JID, Message
from slixmpp.test import SlixTest
import slixmpp.plugins.xep_0421 as xep_0421
from slixmpp.xmlstream import register_stanza_plugin


class TestOccupantId(SlixTest):

    def setUp(self):
        register_stanza_plugin(Message, xep_0421.stanza.OccupantId)

    def testReadOccupantId(self):
        result = """
          <message type='groupchat' from='foo@muc/nick1'>
            <body>Some message</body>
            <occupant-id xmlns='urn:xmpp:occupant-id:0' id='unique-id1'/>
          </message>
        """

        msg = self.Message()
        msg['type'] = 'groupchat'
        msg['from'] = JID('foo@muc/nick1')
        msg['body'] = 'Some message'
        msg['occupant-id']['id'] = 'unique-id1'

        self.check(msg, result)

suite = unittest.TestLoader().loadTestsFromTestCase(TestOccupantId)



import asyncio
import unittest
from uuid import uuid4
from slixmpp import JID
from slixmpp.test.integration import SlixIntegration

UNIQUE = uuid4().hex


class TestMUC(SlixIntegration):

    async def asyncSetUp(self):
        self.mucserver = self.envjid('CI_MUC_SERVER', default='chat.jabberfr.org')
        self.muc = JID('%s@%s' % (UNIQUE, self.mucserver))
        self.add_client(
            self.envjid('CI_ACCOUNT1'),
            self.envstr('CI_ACCOUNT1_PASSWORD'),
        )
        self.add_client(
            self.envjid('CI_ACCOUNT2'),
            self.envstr('CI_ACCOUNT2_PASSWORD'),
        )
        self.register_plugins(['xep_0045'])
        await self.connect_clients()

    async def test_initial_join(self):
        """Check that we can connect to a new muc"""
        self.clients[0]['xep_0045'].join_muc(self.muc, 'client1')
        presence = await self.clients[0].wait_until('muc::%s::got_online' % self.muc)
        self.assertEqual(presence['muc']['affiliation'], 'owner')

    async def test_setup_muc(self):
        """Check that sending the initial room config and affiliation list works"""
        self.clients[0]['xep_0045'].join_muc(self.muc, 'client1')
        presence = await self.clients[0].wait_until('muc::%s::got_online' % self.muc)
        self.assertEqual(presence['muc']['affiliation'], 'owner')
        # Send initial configuration
        config = await self.clients[0]['xep_0045'].get_room_config(self.muc)
        values = config.get_values()
        values['muc#roomconfig_persistentroom'] = False
        values['muc#roomconfig_membersonly'] = True
        config['values'] = values
        config.reply()
        config = await self.clients[0]['xep_0045'].set_room_config(self.muc, config)

        # Send affiliation list including client 2
        await self.clients[0]['xep_0045'].send_affiliation_list(
            self.muc,
            [
                (self.clients[1].boundjid.bare, 'member'),
            ],
        )

    async def test_join_after_config(self):
        """Join a room after being added to the affiliation list"""
        await self.test_setup_muc()
        self.clients[1]['xep_0045'].join_muc(self.muc, 'client2')
        await self.clients[1].wait_until('muc::%s::got_online' % self.muc)

    async def test_nick_change_leave(self):
        """Check that we change nicks and leave properly"""
        await self.test_join_after_config()
        nick = 'coucoucou2'
        new_nick = await self.clients[0]['xep_0045'].set_self_nick(self.muc, nick)
        assert new_nick == nick
        self.clients[0]['xep_0045'].leave_muc(self.muc, 'client1', 'boooring')
        pres = await self.clients[1].wait_until('muc::%s::got_offline' % self.muc)
        self.assertEqual(pres['status'], 'boooring')
        self.assertEqual(pres['type'], 'unavailable')

    async def test_kick(self):
        """Test kicking a user"""
        await self.test_join_after_config()
        await asyncio.gather(
            self.clients[0].wait_until('muc::%s::got_offline' % self.muc),
            self.clients[0]['xep_0045'].set_role(self.muc, 'client2', 'none')
        )


suite = unittest.TestLoader().loadTestsFromTestCase(TestMUC)




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permissio
from slixmpp.plugins.base import register_plugin

from slixmpp.plugins.xep_0313.stanza import Result, MAM, Metadata
from slixmpp.plugins.xep_0313.mam import XEP_0313


register_plugin(XEP_0313)

__all__ = ['XEP_0313', 'Result', 'MAM', 'Metadata']




# Slixmpp: The Slick XMPP Library
# Copyright (C) 2011  Nathanael C. Fritz
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.stanza import Error
from slixmpp.xmlstream import ElementBase, ET, register_stanza_plugin


class PubsubErrorCondition(ElementBase):

    plugin_attrib = 'pubsub'
    interfaces = {'condition', 'unsupported'}
    plugin_attrib_map = {}
    plugin_tag_map = {}
    conditions = {'closed-node', 'configuration-required', 'invalid-jid',
                  'invalid-options', 'invalid-payload', 'invalid-subid',
                  'item-forbidden', 'item-required', 'jid-required',
                  'max-items-exceeded', 'max-nodes-exceeded',
                  'nodeid-required', 'not-in-roster-group',
                  'not-subscribed', 'payload-too-big',
                  'payload-required', 'pending-subscription',
                  'presence-subscription-required', 'subid-required',
                  'too-many-subscriptions', 'unsupported'}
    condition_ns = 'http://jabber.org/protocol/pubsub#errors'

    def setup(self, xml):
        """Don't create XML for the plugin."""
        self.xml = ET.Element('')

    def get_condition(self):
        """Return the condition element's name."""
        for child in self.parent().xml:
            if "{%s}" % self.condition_ns in child.tag:
                cond = child.tag.split('}', 1)[-1]
                if cond in self.conditions:
                    return cond
        return ''

    def set_condition(self, value):
        """
        Set the tag name of the condition element.

        Arguments:
           value -- The tag name of the condition element.
        """
        if value in self.conditions:
            del self['condition']
            cond = ET.Element("{%s}%s" % (self.condition_ns, value))
            self.parent().xml.append(cond)
        return self

    def del_condition(self):
        """Remove the condition element."""
        for child in self.parent().xml:
            if "{%s}" % self.condition_ns in child.tag:
                tag = child.tag.split('}', 1)[-1]
                if tag in self.conditions:
                    self.parent().xml.remove(child)
        return self

    def get_unsupported(self):
        """Return the name of an unsupported feature"""
        xml = self.parent().xml.find('{%s}unsupported' % self.condition_ns)
        if xml is not None:
            return xml.attrib.get('feature', '')
        return ''

    def set_unsupported(self, value):
        """Mark a feature as unsupported"""
        self.del_unsupported()
        xml = ET.Element('{%s}unsupported' % self.condition_ns)
        xml.attrib['feature'] = value
        self.parent().xml.append(xml)

    def del_unsupported(self):
        """Delete an unsupported feature condition."""
        xml = self.parent().xml.find('{%s}unsupported' % self.condition_ns)
        if xml is not None:
            self.parent().xml.remove(xml)


register_stanza_plugin(Error, PubsubErrorCondition)



import logging

from slixmpp import CoroutineCallback, StanzaPath, Iq, register_stanza_plugin
from slixmpp.plugins import BasePlugin
from slixmpp.xmlstream import StanzaBase

from . import stanza


class XEP_0055(BasePlugin):
    """
    XEP-0055: Jabber Search

    The config options are only useful for a "server-side" search feature,
    and if the ``provide_search`` option is set to True.

    API
    ===

    ``search_get_form``: customize the search form content (ie fields)

    ``search_query``: return search results
    """
    name = "xep_0055"
    description = "XEP-0055: Jabber search"
    dependencies = {"xep_0004", "xep_0030"}
    stanza = stanza
    default_config = {
        "form_fields": {"first", "last"},
        "form_instructions": "",
        "form_title": "",
        "provide_search": True
    }

    def plugin_init(self):
        register_stanza_plugin(Iq, stanza.Search)
        register_stanza_plugin(stanza.Search, self.xmpp["xep_0004"].stanza.Form)

        if self.provide_search:
            self.xmpp["xep_0030"].add_feature(stanza.Search.namespace)
            self.xmpp.register_handler(
                CoroutineCallback(
                    "search",
                    StanzaPath("/iq/search"),
                    self._handle_search,
                )
            )
            self.api.register(self._get_form, "search_get_form")
            self.api.register(self._get_results, "search_query")

    async def _handle_search(self, iq: StanzaBase):
        if iq["search"]["form"].get_values():
            reply = await self.api["search_query"](None, None, iq.get_from(), iq)
            reply["search"]["form"]["type"] = "result"
        else:
            reply = await self.api["search_get_form"](None, None, iq.get_from(), iq)
        reply["search"]["form"].add_field(
            "FORM_TYPE", value=stanza.Search.namespace, ftype="hidden"
        )
        reply.send()

    async def _get_form(self, jid, node, ifrom, iq):
        reply = iq.reply()
        form = reply["search"]["form"]
        form["title"] = self.form_title
        form["instructions"] = self.form_instructions
        for field in self.form_fields:
            form.add_field(field)
        return reply

    async def _get_results(self, jid, node, ifrom, iq):
        reply = iq.reply()
        form = reply["search"]["form"]
        form["type"] = "result"

        for field in self.form_fields:
            form.add_reported(field)
        return reply

    def make_search_iq(self, **kwargs):
        iq = self.xmpp.make_iq(itype="set", **kwargs)
        iq["search"]["form"].set_type("submit")
        iq["search"]["form"].add_field(
            "FORM_TYPE", value=stanza.Search.namespace, ftype="hidden"
        )
        return iq


log = logging.getLogger(__name__)




# slixmpp.xmlstream.matcher.xpath
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Part of Slixmpp: The Slick XMPP Library
# :copyright: (c) 2011 Nathanael C. Fritz
# :license: MIT, see LICENSE for more details
from typing import cast
from slixmpp.xmlstream.stanzabase import ET, fix_ns, StanzaBase
from slixmpp.xmlstream.matcher.base import MatcherBase


class MatchXPath(MatcherBase):

    """
    The XPath matcher selects stanzas whose XML contents matches a given
    XPath expression.

    If the value of :data:`IGNORE_NS` is set to ``True``, then XPath
    expressions will be matched without using namespaces.
    """
    _criteria: str

    def __init__(self, criteria: str):
        self._criteria = cast(str, fix_ns(criteria))

    def match(self, xml: StanzaBase) -> bool:
        """
        Compare a stanza's XML contents to an XPath expression.

        If the value of :data:`IGNORE_NS` is set to ``True``, then XPath
        expressions will be matched without using namespaces.

        :param xml: The :class:`~slixmpp.xmlstream.stanzabase.StanzaBase`
                    stanza to compare against.
        """
        real_xml = xml.xml
        x = ET.Element('x')
        x.append(real_xml)

        return x.find(self._criteria) is not None



import unittest
from slixmpp.test.integration import SlixIntegration


class TestConnect(SlixIntegration):
    async def asyncSetUp(self):
        await super().asyncSetUp()
        self.add_client(
            self.envjid('CI_ACCOUNT1'),
            self.envstr('CI_ACCOUNT1_PASSWORD'),
        )
        self.add_client(
            self.envjid('CI_ACCOUNT2'),
            self.envstr('CI_ACCOUNT2_PASSWORD'),
        )
        await self.connect_clients()

    async def test_send_message(self):
        """Make sure we can send and receive messages"""
        msg = self.clients[0].make_message(
            mto=self.clients[1].boundjid, mbody='Msg body',
        )
        msg.send()
        message = await self.clients[1].wait_until('message')
        self.assertEqual(message['body'], msg['body'])


suite = unittest.TestLoader().loadTestsFromTestCase(TestConnect)
